from . import ddm
from . import episodicmemorymechanism
from . import timermechanism

from .ddm import *
from .episodicmemorymechanism import *
from .timermechanism import *

__all__ = list(ddm.__all__)
__all__.extend(episodicmemorymechanism.__all__)
__all__.extend(timermechanism.__all__)



from psyneulink import *

m1 = TransferMechanism(input_ports=["InputPort A", "InputPort B"])
m2 = TransferMechanism()
c = Composition()
c.add_node(m1, required_roles=NodeRole.INPUT)
c.add_node(m2, required_roles=NodeRole.INPUT)
c._analyze_graph()
lvoc = OptimizationControlMechanism(agent_rep=RegressionCFA,
                                    state_features=[m1.input_ports[0], m1.input_ports[1], m2.input_port],
                                    objective_mechanism=ObjectiveMechanism(
                                            monitor=[m1, m2]),
                                    function=GridSearch(max_iterations=1),
                                    control_signals=[(SLOPE, m1), (SLOPE, m2)])
c.add_node(lvoc)
input_dict = {m1: [[1], [1]], m2: [1]}

c.show_graph(show_controller=True)

# c.run(inputs=input_dict)




from . import comparatormechanism
from . import predictionerrormechanism

from .comparatormechanism import *
from .predictionerrormechanism import *

__all__ = list(comparatormechanism.__all__)
__all__.extend(predictionerrormechanism.__all__)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  ParameterPort ******************************************************

"""

Contents
--------

  * `ParameterPort_Overview`
  * `ParameterPort_Creation`
      - `ParameterPort_Specification`
  * `ParameterPort_Structure`
  * `ParameterPort_Execution`
  * `ParameterPort_Class_Reference`

.. _ParameterPort_Overview:

Overview
--------

ParameterPorts belong to either a `Mechanism <Mechanism>` or a `Projection <Projection>`. A ParameterPort is created
to represent each `modulatable parameter <ParameterPort_Modulable_Parameters>` of the `Mechanism
<Mechanism>` or a `Projection <Projection>`, as well as those of the component's `function <Component_Function>` and
any of its secondary functions (e.g. `TransferMechanism.integrator_function`). A
ParameterPort provides the current value of the parameter it represents during any relevant computations, and serves as
an interface for parameter modulation.

A ParameterPort can receive one or more `ControlProjections  <ControlProjection>` and/or `LearningProjections
<LearningProjection>` that modify the value returned by the ParameterPort according to the ParameterPort's
`function <ParameterPort.function>`. The Projections received by a ParameterPort  are listed in its `mod_afferents
<ParameterPort.mod_afferents>` attribute.

When the Mechanism or Projection to which a ParameterPort belongs executes, that component and its function use the
ParameterPort's value -- not the parameter attribute's value -- for any computation. A ParameterPort's corresponding
attribute on the Mechanism, Projection, or Function to which it belongs (i.e. MyTransferMech.function.gain),
stores the "base value" of that parameter. The base value of a parameter is the variable of the ParameterPort's
function. The base value can be viewed or changed at any time through this attribute.

The ParameterPort value is available on the ParameterPort itself, as well as the mod_name attribute of the Mechanism
or Projection to which it belongs (i.e. MyTransferMech.mod_gain would return the value of the "gain" ParameterPort
of the MyTransferMech mechanism.)

.. note::
    Either of these options for looking up the value of the ParameterPort will return the ParameterPort value that
    was used during the most recent execution. This means that if the value of MyTransferMech.function.gain (the
    base value) is updated after execution #1, the base value will change immediately, but the ParameterPort value (and
    MyTransferMech.mod_gain) will not be computed again until execution #2.

    As a result, if either MyTransferMech.mod_gain or MyTransferMech.parameter_ports["gain"].value is viewed in between
    execution #1 and execution #2, it will return the gain ParameterPort value that was used during execution 1.

.. _ParameterPort_Creation:

Creating a ParameterPort
-------------------------

ParameterPorts are created automatically when the `Mechanism <Mechanism>` or `Projection <Projection>` to which they
belong is created.  The `owner <Port.owner>` of a ParameterPort must be a `Mechanism <Mechanism>` or `MappingProjection`
(the initialization of a ParameterPort cannot be `deferred <Port_Deferred_Initialization>`). One ParameterPort is
created for each modulable Parameter of its owner, as well as for each modulable Parameter of the owner's
`function <Component.function>` or secondary functions (modulable
Parameters of a Component
are listed in its `Parameters` class, and have the attribute
`modulable <Parameter.modulable>` set to True.)
Each ParameterPort is created using the value specified for the corresponding parameter, as described below.
The ParameterPorts for the parameters of a Mechanism or Projection are listed in its `parameter_ports
<Mechanism_Base.parameter_ports>` attribute.

COMMENT:
    FOR DEVELOPERS: The instantiation of ParameterPorts for all of the `user_params` of a Component can be
                    suppressed if a *PARAMETER_PORTS* entry is included and set to `NotImplemented` in the
                    default of its class Parameters;  the instantiation of a ParameterPort
                    for an individual parameter in user_params can be suppressed by including it in
                    exclude_from_parameter_ports for the class (or one of its parent classes)
                    (see LearningProjection and EVCControlMechanism for examples, and `note
                    <ParameterPorts_Suppression>` below for additional information about how
                    to suppress creation of a ParameterPort for individual parameters.  This should be done
                    for any parameter than can take a value or a string that is a keyword as its specification
                    (i.e., of the arg for the parameter in the Component's constructor) but should not have a
                    ParameterPort (e.g., input_port and output_port), as otherwise the
                    specification will be interpreted as a numeric parameter (in the case of a value) or
                    a parameter of the keyword's type, a ParameterPort will be created, and then it's value,
                    rather than the parameter's actual value, will be returned when the parameter is accessed
                    using "dot notation" (this is because the getter for an attribute's property first checks
                    to see if there is a ParameterPort for that attribute and, if so, returns the value of the
                    ParameterPort).
COMMENT

.. _ParameterPort_Specification:

*Specifying Parameters*
~~~~~~~~~~~~~~~~~~~~~~~

Parameters can be specified in one of several places:

    * In the **argument** of the constructor for the `Component <Component>` to which the parameter belongs
      (see `Component_Structural_Attributes` for additional details).
    ..
    * In a *parameter specification dictionary* assigned to the **params** argument in the constructor for the
      Component to which the parameter belongs, or any place else the value of a parameter can be specified.
      The entry for each parameter must use the name of the parameter (or a corresponding keyword) as its key,
      and the parameter's specification as its value (see `examples <ParameterPort_Specification_Examples>` below).
      Parameters for a Component's `function <Component.function>` can be specified in an entry with the key
      *FUNCTION_PARAMS*, and a value that is itself a parameter specification dictionary containing an entry for
      each of the function's parameters to be specified.  When a value is assigned to a parameter in a specification
      dictionary, it overrides any value assigned to the argument for the parameter in the Component's constructor.
    ..
    * By direct assignment to the Component's attribute for the parameter
      (see `below <ParameterPort_Modulable_Parameters>`).
    ..
    ..
    * In the **runtime_params** argument of a call to a Composition's `Run` method

.. _ParameterPort_Value_Specification:

The specification of the initial value of a parameter can take any of the following forms:

    .. _ParameterPort_Value_Assignment:

    * **Value** -- this must be a valid value for the parameter. It creates a default ParameterPort,
      assigns the parameter's default value as the ParameterPort's `value <ParameterPort.value>`,
      and assigns the parameter's name as the name of the ParameterPort.

    * **ParameterPort reference** -- this must refer to an existing **ParameterPort** object; its name must be the
      name of a parameter of the owner or of the owner's `function <Component.function>`, and its value must be a valid
      one for the parameter.

      .. note::
          This capability is provided for generality and potential
          future use, but its current use is not advised.

    .. _ParameterPort_Modulatory_Specification:

    * **Modulatory specification** -- this can be an existing `ControlSignal` or `ControlProjection`,
      a `LearningSignal` or `LearningProjection`, a constructor or the class name for any of these, or the
      keywords *CONTROL*, *CONTROL_PROJECTION*, *LEARNING*, or *LEARNING_PROJECTION*.  Any of these create a default
      ParameterPort, assign the parameter's default value as the ParameterPort's `value <ParameterPort.value>`,
      and assign the parameter's name as the name of the ParameterPort.  They also create and/or assign the
      corresponding ModulatorySignal and ModulatoryProjection, and assign the ParameterPort as the
      ModulatoryProjection's `receiver <Projection_Base.receiver>`. If the ModulatorySignal and/or
      ModulatoryProjection already exist, their value(s) must be valid one(s) for the parameter.  Note that only
      Control and Learning Modulatory components can be assigned to a ParameterPort (Gating components cannot be
      used -- they can only be assigned to `InputPorts <InputPort>` and `OutputPorts <OutputPort>`).

    .. _ParameterPort_Tuple_Specification:

    * **2-item tuple:** *(<value>, <Modulatory specification>)* -- this creates a default ParameterPort, uses the value
      specification (1st item) as parameter's `value assignment <ParameterPort_Value_Assignment>`, and assigns the
      parameter's name as the name of the ParameterPort.  The Modulatory specification (2nd item) is used as the
      ParameterPort's `modulatory assignment <ParameterPort_Modulatory_Specification>`, and the ParameterPort
      is assigned as the `receiver <Projection_Base.receiver>` for the corresponding `ModulatoryProjection
      <ModulatoryProjection>`.

      .. note::
          Currently, the `function <Component.function>` of a Component, although it can be specified as a parameter
          value, cannot be assigned a `ModulatorySignal <ModulatorySignal>` or modified in the **runtime_params**
          argument of a call to a Mechanism's `execute <Mechanism_Base.execute>` method. This may change in the future.

The value specified for a parameter (either explicitly or by default) is assigned to an attribute of the Component or
of the Component's `function <Mechanism_Base.function>` to which the parameter belongs.  The attribute has the same
name as the parameter, and can be referenced using standard Python attribute ("dot") notation;  for example, the value
of a parameter named *param* is assigned to an attribute named ``param`` that can be referenced as
``my_component.param``). The parameter's value is assigned as the **default value** for the ParameterPort.

.. _ParameterPorts_Suppression:

.. note::
   If the value of a parameter is specified as `NotImplemented`, or any non-numeric value that is not one of those
   listed above, then no ParameterPort is created and the parameter cannot be modified by a `ModulatorySignal
   <ModulatorySignal>` or in the **runtime_params** argument of a call to a Mechanism's `execute
   <Mechanism_Base.execute>` method.

.. _ParameterPort_Specification_Examples:

*Examples*
~~~~~~~~~~

In the following example, a Mechanism is created by specifying two of its parameters, as well as its
`function <Component.function>` and two of that function's parameters, each using a different specification format::

    >>> import psyneulink as pnl
    >>> my_mechanism = pnl.RecurrentTransferMechanism(
    ...                         input_shapes=5,
    ...                         noise=pnl.ControlSignal(),
    ...                         function=pnl.Logistic(
    ...                                         gain=(0.5, pnl.ControlSignal),
    ...                                         bias=(1.0, pnl.ControlSignal(modulation=pnl.ADDITIVE))))

COMMENT:
    If assigning a default ControlSignal makes the noise value the same as the
    default noise value, why are we using a ControlSignal here??
COMMENT

The first argument of the constructor for the Mechanism specifies its `input_shapes <Component.input_shapes>` parameter by
directly assigning a value to it.  The second specifies the `noise <RecurrentTransferMechanism.noise>` parameter
by assigning a default `ControlSignal`;  this will use the default value of the
`noise <RecurrentTransferMechanism.noise>` attribute.  The **function** argument is specified using the constructor for
a `Logistic` function, that specifies two of its parameters.  The `gain <Logistic.gain>` parameter
is specified using a tuple, the first item of which is the value to be assigned, and the second specifies
a default `ControlSignal`.  The `bias <Logistic.bias>` parameter is also specified using a tuple,
in this case with a constructor for the ControlSignal that specifies its `modulation <ModulatorySignal.modulation>`
parameter.

In the following example, a `MappingProjection` is created, and its
`matrix <MappingProjection.MappingProjection.matrix>` parameter is assigned a random weight matrix (using a
`matrix keyword <Matrix_Keywords>`) and `LearningSignal`::

    >>> my_input_mechanism = pnl.TransferMechanism()
    >>> my_output_mechanism = pnl.TransferMechanism()
    >>> my_mapping_projection = pnl.MappingProjection(sender=my_input_mechanism,
    ...                                               receiver=my_output_mechanism,
    ...                                               matrix=(pnl.RANDOM_CONNECTIVITY_MATRIX,
    ...                                                       pnl.LearningSignal))

.. note::
   The `matrix <MappingProjection.MappingProjection.matrix>` parameter belongs to the MappingProjection's `function
   <Projection_Base.function>`;  however, since it has only one standard function, its arguments are available in the
   constructor for the Projection (see `here <User_Modifiable_Parameters>` for a more detailed explanation).

The example below shows how to specify the parameters in the first example using a parameter specification dictionary::

    >>> my_mechanism = pnl.RecurrentTransferMechanism(
    ...                      noise=5,
    ...                      params={pnl.NOISE: pnl.CONTROL,
    ...                              pnl.FUNCTION: pnl.Logistic,
    ...                              pnl.FUNCTION_PARAMS:{
    ...                                     pnl.GAIN:(0.5,pnl.ControlSignal),
    ...                                     pnl.BIAS:(1.0,pnl.ControlSignal(modulation=pnl.ADDITIVE))}})

There are several things to note here.

First, the parameter specification dictionary must be assigned to the **params** argument of the constructor. Note that
if the parameter is specified in a parameter specification dictionary, the key for the parameter must be a string that
is the same as the name of parameter (i.e., identical to how it appears as an arg in the constructor; as is shown
for **noise** in the example), or using a keyword that resolves to such a string (as shown for *NOISE* in the
example).

Second, both methods for specifying a parameter -- directly in an argument for the parameter, or in an entry of a
parameter specification dictionary -- can be used within the same constructor.

If a particular parameter is specified in both ways (as is the case for **noise** in the example), the value in the
parameter specification dictionary takes priority (i.e., it is the value that will be assigned to the parameter).

Finally, the keyword *FUNCTION_PARAMS* can be used in a parameter specification dictionary to specify
parameters of the Component's `function <Component.function>`, as shown for the **gain** and **bias** parameters of
the Logistic function in the example.

The example below shows how to access ParameterPort values vs base values, and demonstrates their differences:

    >>> my_transfer_mechanism = pnl.TransferMechanism(
    ...                      noise=5.0,
    ...                      function=pnl.Linear(slope=2.0))
    >>> assert my_transfer_mechanism.noise.base == 5.0
    >>> assert my_transfer_mechanism.mod_noise == [5.0]
    >>> assert my_transfer_mechanism.function.slope.base == 2.0
    >>> assert my_transfer_mechanism.mod_slope == [2.0]

Notice that the noise attribute, which stores the base value for the noise ParameterPort of my_transfer_mechanism, is
on my_transfer_mechanism, while the slope attribute, which stores the base value for the slope ParameterPort of
my_transfer_mechanism, is on my_transfer_mechanism's function. However, mod_noise and mod_slope are both properties on
my_transfer_mechanism.

    >>> my_transfer_mechanism.noise.base = 4.0
    >>> my_transfer_mechanism.function.slope.base = 1.0
    >>> assert my_transfer_mechanism.noise.base == 4.0
    >>> assert my_transfer_mechanism.mod_noise == [5.0]
    >>> assert my_transfer_mechanism.function.slope.base == 1.0
    >>> assert my_transfer_mechanism.mod_slope == [2.0]

When the base values of noise and slope are updated, we can inspect these attributes immediately and observe that they
have changed. We do not observe a change in mod_noise or mod_slope because the ParameterPort value will not update
until the mechanism executes.

    >>> my_transfer_mechanism.execute([10.0])
    array([[14.]])
    >>> assert my_transfer_mechanism.noise.base == 4.0
    >>> assert my_transfer_mechanism.mod_noise == [4.0]
    >>> assert my_transfer_mechanism.function.slope.base == 1.0
    >>> assert my_transfer_mechanism.mod_slope == 1.0

Now that the mechanism has executed, we can see that each ParameterPort evaluated its function with the base value,
producing a modulated noise value of 4.0 and a modulated slope value of 1.0. These values were used by
my_transfer_mechanism and its Linear function when the mechanism executed.

.. _ParameterPort_Structure:

Structure
---------

Every ParameterPort is owned by a `Mechanism <Mechanism>` or `MappingProjection`. It can receive one or more
`ControlProjections <ControlProjection>` or `LearningProjections <LearningProjection>`, that are listed in its
`mod_afferents <ParameterPort.mod_afferents>` attribute.  A ParameterPort cannot receive
`PathwayProjections <PathwayProjection>` or `GatingProjections <GatingProjection>`.  When the ParameterPort is
updated (i.e., its owner is executed), it uses the values of its ControlProjections and LearningProjections to
determine whether and how to modify its parameter's attribute value, which is then assigned as the ParameterPort's
`value <ParameterPort.value>` (see `ParameterPort_Execution` for addition details). ParameterPorts have the
following core attributes:

* `variable <ParameterPort.variable>` - the parameter's attribute value; that is, the value assigned to the
  attribute for the parameter of the ParameterPort's owner;  it can be thought of as the parameter's "base" value.
  It is used by its `function <ParameterPort.function>` to determine the ParameterPort's
  `value <ParameterPort.value>`.  It must match the format (the number and type of elements) of the parameter's
  attribute value.

* `mod_afferents <ParameterPort.mod_afferents>` - lists the `ModulatoryProjections <ModulationProjection>` received
  by the ParameterPort.  These specify either modify the ParameterPort's `function <ParameterPort.function>`, or
  directly assign the `value <ParameterPort.value>` of the ParameterPort itself (see `ModulatorySignals_Modulation`).

* `function <ParameterPort.function>` - takes the parameter's attribute value as its input, modifies it under the
  influence of any `ModulatoryProjections` it receives (listed in `mod_afferents <ParameterPort.mod_afferents>`,
  and assigns the result as the ParameterPort's `value <ParameterPort.value>` which is used as the parameter's
  "actual" value.

* `value <ParameterPort.value>` - the result of `function <ParameterPort.function>`; used by the ParameterPort's
  owner as the value of the parameter for which the ParameterPort is responsible.

.. _ParameterPort_Modulable_Parameters:

All of the modulable parameters of a Component -- that is, for which it has ParameterPorts --
are listed in its `Parameters` class, and have the attribute
`modulable <Parameter.modulable>` set to True.  The
ParameterPorts for a Mechanism or Projection are listed in its `parameter_ports <Mechanism_Base.parameter_ports>`
attribute, which is also read-only.

An initial value can be assigned to a parameter in the corresponding argument of the constructor for the Component
(see `above <ParameterPort_Value_Specification>`.  Parameter values can also be modified by a assigning a value to
the corresponding attribute.
The parameters of a Component's function can be modified by assigning a value to the corresponding attribute of the
Component's `function <Component.function>` attribute (e.g., ``myMechanism.function.my_parameter``).
See `Mechanism_ParameterPorts` for additional information.


.. _ParameterPort_Execution:

Execution
---------

A ParameterPort cannot be executed directly.  It is executed when the Component to which it belongs is executed.
When this occurs, the ParameterPort executes any `ModulatoryProjections` it receives, the values of which
modulate parameters of the ParameterPort's `function <ParameterPort.function>`.  The ParameterPort then calls
its `function <ParameterPort.function>` and the result is assigned as its `value <ParameterPort.value>`.  The
ParameterPort's `value <ParameterPort.value>` is used as the value of the corresponding parameter by the Component,
or by its own `function <Component.function>`.

.. note::
   It is important to note the distinction between the `function <ParameterPort.function>` of a ParameterPort,
   and the `function <Component.function>` of the Component to which it belongs. The former is used to determine the
   value of a parameter used by the latter (see `figure <ModulatorySignal_Anatomy_Figure>`, and `Port_Execution` for
   additional details).

.. _ParameterPort_Class_Reference:

Class Reference
---------------

"""

import collections
import inspect
import operator
import types
import warnings
from copy import deepcopy

import numpy as np
from beartype import beartype

from psyneulink._typing import Optional

from psyneulink.core.components.component import Component, parameter_keywords
from psyneulink.core.components.functions.function import FunctionError, get_param_value_for_keyword
from psyneulink.core.components.ports.modulatorysignals.modulatorysignal import ModulatorySignal
from psyneulink.core.components.ports.port import PortError, Port_Base, _instantiate_port, port_type_keywords
from psyneulink.core.components.shellclasses import Mechanism, Projection, Function
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import \
    CONTEXT, CONTROL_PROJECTION, CONTROL_SIGNAL, CONTROL_SIGNALS, FUNCTION, FUNCTION_PARAMS, \
    LEARNING_SIGNAL, LEARNING_SIGNALS, MECHANISM, NAME, PARAMETER_PORT, PARAMETER_PORT_PARAMS, PATHWAY_PROJECTION, \
    PROJECTION, PROJECTIONS, PROJECTION_TYPE, REFERENCE_VALUE, SENDER, VALUE
from psyneulink.core.globals.parameters import ParameterAlias, SharedParameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel
from psyneulink.core.globals.utilities \
    import ContentAddressableList, ReadOnlyOrderedDict, is_iterable, is_numeric, is_value_spec, iscompatible, \
    is_instance_or_subclass, gen_friendly_comma_str

__all__ = [
    'ParameterPort', 'ParameterPortError', 'port_type_keywords',
]

port_type_keywords = port_type_keywords.update({PARAMETER_PORT})


class ParameterPortList(ContentAddressableList):

    separator = '-'
    legal_key_type_strings = ContentAddressableList.legal_key_type_strings + ['Parameter']
    _owner_port_suffix = 'self'

    def __init__(
        self,
        component_type,
        key=None,
        list=None,
        name=None,
        owner=None,
        **kwargs
    ):
        # cache, Parameter keys added when creating Ports, others upon lookup
        self.parameter_mapping = {}
        self.owner = owner

        super().__init__(component_type, key, list, name, **kwargs)

    def __contains__(self, item):
        try:
            return super().__contains__(item)
        except ParameterPortError:
            return False

    def __getitem__(self, key):
        try:
            return self.parameter_mapping[key]
        except KeyError:
            pass

        # is a Parameter
        try:
            final_source = key.final_source
        except AttributeError:
            pass
        else:
            try:
                return self.parameter_mapping[final_source]
            except KeyError as e:
                try:
                    raise ParameterPortError(
                        f'No ParameterPort corresponds to {key._owner._owner}'
                        f'.parameters.{key.name}'
                    ) from None
                except AttributeError:
                    raise e from None

        try:
            return super().__getitem__(key)
        except TypeError as e:
            # ContentAddressableList throws TypeError when key/index lookup fails
            names = self._get_possible_port_names(key)
            possible_ports = set()
            for name in names:
                try:
                    r = super().__getitem__(name)
                    possible_ports.add(r)
                except TypeError:
                    pass
            if len(possible_ports) == 0:
                raise e from None
            elif len(possible_ports) == 1:
                res = next(iter(possible_ports))
            else:
                raise ParameterPortError(
                    f'Multiple ParameterPorts for {key} exist. Did you want'
                    f' {gen_friendly_comma_str(sorted([p.name for p in possible_ports]))}?'
                ) from None

        if res is not None:
            self.parameter_mapping[key] = res

        return res

    def __delitem__(self, key):
        main_port = self[key]
        rem_mapping_keys = set()

        for m, port in self.parameter_mapping.items():
            if port is main_port:
                rem_mapping_keys.add(m)

        for m in rem_mapping_keys:
            del self.parameter_mapping[m]

        del self.data[self.data.index(main_port)]

    def _get_possible_port_names(self, param_name):
        """
            Returns:
                a list of possible parameter port names to check if
                *param_name* is actually an alias or alias-with-suffix
                (e.g. "leak" is an alias of "integration_rate", and
                "leak__integrator_function" should refer to
                "integration_rate__integrator_function")
        """
        unsuffixed_name = ParameterPortList._get_base_name(param_name)
        if unsuffixed_name == param_name:
            # all possible function-suffixed names
            names = sorted(
                [
                    p.name for p in self.owner.parameters
                    if is_instance_or_subclass(p.default_value, Function)
                ]
                + [self._owner_port_suffix]
            )
            # put 'function' at beginning
            try:
                function_index = names.index(FUNCTION)
                names = (
                    [names[function_index]]
                    + names[0:function_index]
                    + names[function_index + 1:]
                )
            except ValueError:
                pass

            names = [self._get_explicit_name(param_name, name) for name in names]
        else:
            names = []

        # try to get a Parameter that corresponds to param_name, which
        # can have a "shared parameter suffix" that disambiguates which
        # desired port it refers to if there are multiple
        try:
            param = getattr(self.owner.parameters, param_name)
        except AttributeError:
            try:
                param = getattr(self.owner.parameters, unsuffixed_name)
            except AttributeError:
                return names

        # if it's a shared parameter with identical name, there are no
        # other aliases we need to add
        try:
            source_name = param.source.name
        except AttributeError:
            return names

        if source_name != param.name:
            if unsuffixed_name == param_name:
                # basic alias, e.g. "leak" -> "integration_rate"
                names.append(source_name)
            else:
                # alias with suffix, e.g. "leak__function"
                # -> "integration_rate__function"
                suffix = ParameterPortList._get_suffix(param_name)
                names.append(
                    ParameterPortList._get_explicit_name(source_name, suffix)
                )

            if isinstance(param, ParameterAlias):
                # alias to another alias or a shared parameter
                # e.g. leak -> integration_rate -> rate
                names.extend(self._get_possible_port_names(source_name))
            else:
                # e.g. integration_rate__integrator_function
                # -> rate__integrator_function
                names.append(
                    ParameterPortList._get_explicit_name(
                        source_name,
                        param.attribute_name
                    )
                )

        return names

    @classmethod
    def _get_explicit_name(cls, port_name, parameter_names=None):
        if isinstance(parameter_names, str):
            parameter_names = [parameter_names]

        if parameter_names is not None:
            suffix = cls.separator + cls.separator.join(p for p in parameter_names)
        else:
            suffix = ''

        return f'{port_name}{suffix}'

    @classmethod
    def _get_base_name(cls, explicit_name):
        try:
            return explicit_name.split(cls.separator)[0]
        except IndexError:
            return explicit_name

    @classmethod
    def _get_suffix(cls, explicit_name):
        try:
            return cls.separator.join(explicit_name.split(cls.separator)[1:])
        except IndexError:
            return ''


class ParameterPortError(PortError):
    pass


class ParameterPort(Port_Base):
    """
    ParameterPort(                                           \
        owner,                                               \
        reference_value=None                                 \
        function=LinearCombination(operation=PRODUCT),       \


    Subclass of `Port <Port>` that represents and possibly modifies the parameter of a `Mechanism <Mechanism>`,
    `Projection <Projection>`, or its `Function`. See `Port_Class_Reference` for additional arguments and attributes.

    COMMENT:
    PortRegistry
    -------------
        All ParameterPorts are registered in PortRegistry, which maintains an entry for the subclass,
        a count for all instances of it, and a dictionary of those instances

    COMMENT

    Arguments
    ---------

    owner : Mechanism or MappingProjection
        the `Mechanism <Mechanism>` or `MappingProjection` to which to which the ParameterPort belongs; it must be
        specified or determinable from the context in which the ParameterPort is created (the initialization of a
        ParameterPort cannot be `deferred <Port_Deferred_Initialization>`. The owner of a ParameterPort
        for the parameter of a `function <Component.function>` should be specified as the Mechanism or Projection to
        which the function belongs.

    reference_value : number, list or np.ndarray
        specifies the default value of the parameter for which the ParameterPort is responsible.

    variable : number, list or np.ndarray
        specifies the parameter's initial value and attribute value — that is, the value of the attribute of the
        ParameterPort's owner or its `function <Component.function>` assigned to the parameter.

    function : Function or method : default LinearCombination(operation=SUM)
        specifies the function used to convert the parameter's attribute value (same as the ParameterPort's
        `variable <ParameterPort.variable>`) to the ParameterPort's `value <ParameterPort.value>`.


    Attributes
    ----------

    mod_afferents : Optional[List[Projection]]
        a list of the `ModulatoryProjection <ModulatoryProjection>` that project to the ParameterPort (i.e.,
        for which it is a `receiver <Projection_Base.receiver>`); these can be `ControlProjection(s)
        <ControlProjection>` and/or `LearningProjection(s) <LearningProjection>`, but not `GatingProjection
        <GatingProjection>`.  The `value <ModulatoryProjection_Base.value>` of each must match the format
        (number and types of elements) of the ParameterPort's `variable <ParameterPort.variable>`.

    variable : number, list or np.ndarray
        the parameter's attribute value — that is, the value of the attribute of the
        ParameterPort's owner or its `function <Component.function>` assigned to the parameter.

    function : Function : default Linear
        converts the parameter's attribute value (same as the ParameterPort's `variable <ParameterPort.variable>`)
        to the ParameterPort's `value <ParameterPort.value>`, under the influence of any
        `ModulatoryProjections <ModulatoryProjection>` received by the ParameterPort (and listed in its
        `mod_afferents <ParameterPort.mod_afferents>` attribute.  The result is assigned as the ParameterPort's
        `value <ParameterPort>`.

    value : number, List[number] or np.ndarray
        the result returned by the ParameterPort's `function <ParameterPort.function>`, and used by the
        ParameterPort's owner or its `function <Component.function>` as the value of the parameter for which the
        ParmeterPort is responsible.  Note that this is not necessarily the same as the parameter's attribute value
        (that is, the value of the owner's attribute for the parameter), since the ParameterPort's
        `function <ParameterPort.function>` may modify the latter under the influence of its
        `mod_afferents <ParameterPort.mod_afferents>`.

    """

    #region CLASS ATTRIBUTES

    componentType = PARAMETER_PORT
    paramsType = PARAMETER_PORT_PARAMS

    portAttributes = Port_Base.portAttributes

    connectsWith = [CONTROL_SIGNAL, LEARNING_SIGNAL]
    connectsWithAttribute = [CONTROL_SIGNALS, LEARNING_SIGNALS]
    projectionSocket = SENDER
    modulators = [CONTROL_SIGNAL, LEARNING_SIGNAL]
    canReceive = modulators
    projection_type = CONTROL_PROJECTION

    classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TYPE_DEFAULT_PREFERENCES
    # Note: only need to specify setting;  level will be assigned to TYPE automatically
    # classPreferences = {
    #     PREFERENCE_SET_NAME: 'ParameterPortCustomClassPreferences',
    #     PREFERENCE_KEYWORD<pref>: <setting>...}

    #endregion

    @check_user_specified
    @beartype
    def __init__(self,
                 owner,
                 reference_value=None,
                 variable=None,
                 input_shapes=None,
                 function=None,
                 projections=None,
                 params=None,
                 name=None,
                 parameter_name=None,
                 prefs:   Optional[ValidPrefSet] = None,
                 **kwargs):

        # If context is not COMPONENT or CONSTRUCTOR, raise exception
        context = kwargs.pop(CONTEXT, None)
        if context is None:
            raise ParameterPortError(f"Contructor for {self.__class__.__name__} cannot be called directly"
                                      f"(context: {context}")

        # FIX: UPDATED TO INCLUDE LEARNING [CHANGE THIS TO INTEGRATOR FUNCTION??]
        # # Reassign default for MATRIX param of MappingProjection
        # if isinstance(owner, MappingProjection) and name is MATRIX:
        #     function = LinearCombination(operation=SUM)

        self.reference_value = reference_value

        # Validate sender (as variable) and params
        # Note: pass name of Mechanism (to override assignment of componentName in super.__init__)
        super(ParameterPort, self).__init__(owner,
                                            variable=variable,
                                            input_shapes=input_shapes,
                                            projections=projections,
                                            function=function,
                                            params=params,
                                            name=name,
                                            prefs=prefs,
                                            context=context)

    def _validate_against_reference_value(self, reference_value):
        """Validate that value of the Port is compatible with the reference_value

        reference_value is the value of the parameter to which the ParameterPort is assigned
        """
        if reference_value is not None and not iscompatible(np.squeeze(reference_value), np.squeeze(self.defaults.value)):
            iscompatible(np.squeeze(reference_value), np.squeeze(self.defaults.value))
            name = self.name or ""
            raise ParameterPortError("Value specified for {} {} of {} ({}) is not compatible "
                                      "with its expected format ({})".
                                      format(name, self.componentName, self.owner.name, self.defaults.value, reference_value))

    def _instantiate_projections(self, projections, context=None):
        """Instantiate Projections specified in PROJECTIONS entry of params arg of Port's constructor

        Disallow any PathwayProjections
        Call _instantiate_projections_to_port to assign ModulatoryProjections to .mod_afferents

        """

        # MODIFIED 7/8/17
        # FIX:  THIS SHOULD ALSO LOOK FOR OTHER FORMS OF SPECIFICATION
        # FIX:  OF A PathwayProjection (E.G., TARGET PORT OR MECHANISM)

        from psyneulink.core.components.projections.pathway.pathwayprojection import PathwayProjection_Base
        pathway_projections = [proj for proj in projections if isinstance(proj, PathwayProjection_Base)]
        if pathway_projections:
            pathway_proj_names = []
            for proj in pathway_projections:
                pathway_proj_names.append(proj.name + ' ')
            raise PortError("{} not allowed for {}: {}".
                             format(PathwayProjection_Base.__self__.__name__,
                                    self.__class__.__name__,
                                    pathway_proj_names))

        self._instantiate_projections_to_port(projections=projections, context=context)

    def _check_for_duplicate_projections(self, projection):
        """Check if projection is redundant with one in mod_afferents of ParameterPort

        Check for any instantiated projection in mod_afferents with the same sender as projection
        or one in deferred_init status with sender specification that is the same type as projection.

        Returns redundant Projection if found, otherwise False.
        """

        duplicate = next(iter([proj for proj in self.mod_afferents
                               if ((proj.sender == projection.sender and proj != projection)
                                   or (proj.initialization_status == ContextFlags.DEFERRED_INIT
                                       and proj._init_args[SENDER] == type(projection.sender)))]), None)
        if duplicate and self.verbosePref or self.owner.verbosePref:
            from psyneulink.core.components.projections.projection import Projection
            warnings.warn(f'{Projection.__name__} from {projection.sender.name}  {projection.sender.__class__.__name__}'
                          f' of {projection.sender.owner.name} to {self.name} {self.__class__.__name__} of '
                          f'{self.owner.name} already exists; will ignore additional one specified ({projection.name}).')
        return duplicate

    def _get_all_afferents(self):
        return self.mod_afferents

    def _get_all_projections(self):
        return self.mod_afferents

    @beartype
    def _parse_port_specific_specs(self, owner, port_dict, port_specific_spec, context=None):
        """Get connections specified in a ParameterPort specification tuple

        Tuple specification can be:
            (port_spec, projections)
        Assumes that port_spec has already been extracted and used by _parse_port_spec

        Returns params dict with PROJECTIONS entries if any of these was specified.

        """
        from psyneulink.core.components.projections.projection import _parse_connection_specs, _is_projection_spec

        params_dict = {}
        port_spec = port_specific_spec

        if isinstance(port_specific_spec, dict):
            return None, port_specific_spec

        elif isinstance(port_specific_spec, tuple):

            tuple_spec = port_specific_spec

            # GET PORT_SPEC (PARAM VALUE) AND ASSIGN PROJECTIONS_SPEC **********************************************

            # 2-item tuple specification
            if len(tuple_spec) == 2:

                # 1st item is a value, so treat as Port spec (and return to _parse_port_spec to be parsed)
                #   and treat 2nd item as Projection specification
                if is_numeric(tuple_spec[0]):
                    port_spec = tuple_spec[0]
                    reference_value = port_dict[REFERENCE_VALUE]
                    # Assign value so sender_dim is skipped below
                    # (actual assignment is made in _parse_port_spec)
                    if reference_value is None:
                        port_dict[REFERENCE_VALUE]=port_spec
                    elif not iscompatible(port_spec, reference_value):
                        raise PortError("Value in first item of 2-item tuple specification for {} of {} ({}) "
                                         "is not compatible with its {} ({})".
                                         format(ParameterPort.__name__, owner.name, port_spec,
                                                REFERENCE_VALUE, reference_value))
                    projections_spec = tuple_spec[1]

                elif _is_projection_spec(tuple_spec[0], include_matrix_spec=True):
                    port_spec, projections_spec = tuple_spec

                # Tuple is Projection specification that is used to specify the Port,
                else:
                    # return None in port_spec to suppress further, recursive parsing of it in _parse_port_spec
                    port_spec = None
                    if tuple_spec[0] != self:
                        # If 1st item is not the current port (self), treat as part of the projection specification
                        projections_spec = tuple_spec
                    else:
                        # Otherwise, just use 2nd item as projection spec
                        port_spec = None
                        projections_spec = tuple_spec[1]

            # 3- or 4-item tuple specification
            elif len(tuple_spec) in {3,4}:
                # Tuple is projection specification that is used to specify the Port,
                #    so return None in port_spec to suppress further, recursive parsing of it in _parse_port_spec
                port_spec = None
                # Reduce to 2-item tuple Projection specification
                projection_item = tuple_spec[3] if len(tuple_spec)==4 else None
                projections_spec = (tuple_spec[0],projection_item)

            # GET PROJECTIONS IF SPECIFIED *************************************************************************

            try:
                projections_spec
            except UnboundLocalError:
                pass
            else:
                try:
                    params_dict[PROJECTIONS] = _parse_connection_specs(self,
                                                                       owner=owner,
                                                                       connections=projections_spec)

                    # Parse the value of all of the Projections to get/validate parameter value
                    from psyneulink.core.components.projections.modulatory.controlprojection import ControlProjection
                    from psyneulink.core.components.projections.modulatory.learningprojection import LearningProjection

                    for projection_spec in params_dict[PROJECTIONS]:
                        if port_dict[REFERENCE_VALUE] is None:
                            # FIX: - PUTTING THIS HERE IS A HACK...
                            # FIX:     MOVE TO _parse_port_spec UNDER PROCESSING OF ProjectionTuple SPEC
                            # FIX:     USING _get_port_for_socket
                            # from psyneulink.core.components.projections.projection import _parse_projection_spec

                            # defaults.value?
                            mod_signal_value = projection_spec.port.value \
                                if isinstance(projection_spec.port, Port_Base) else None

                            mod_projection = projection_spec.projection
                            if isinstance(mod_projection, dict):
                                if mod_projection[PROJECTION_TYPE] not in {ControlProjection, LearningProjection}:
                                    raise ParameterPortError("PROGRAM ERROR: {} other than {} or {} ({}) found "
                                                              "in specification tuple for {} param of {}".
                                                              format(Projection.__name__,
                                                                     ControlProjection.__name__,
                                                                     LearningProjection.__name__,
                                                                     mod_projection, port_dict[NAME], owner.name))
                                elif VALUE in mod_projection:
                                    mod_proj_value = mod_projection[VALUE]
                                else:
                                    mod_proj_value = None
                            elif isinstance(mod_projection, Projection):
                                if not isinstance(mod_projection, (ControlProjection, LearningProjection)):
                                    raise ParameterPortError("PROGRAM ERROR: {} other than {} or {} ({}) found "
                                                              "in specification tuple for {} param of {}".
                                                              format(Projection.__name__,
                                                                     ControlProjection.__name__,
                                                                     LearningProjection.__name__,
                                                                     mod_projection, port_dict[NAME], owner.name))
                                elif mod_projection.initialization_status == ContextFlags.DEFERRED_INIT:
                                    continue
                                mod_proj_value = mod_projection.defaults.value
                            else:
                                raise ParameterPortError("Unrecognized Projection specification for {} of {} ({})".
                                                      format(self.name, owner.name, projection_spec))

                            # FIX: 11/25/17 THIS IS A MESS:  CHECK WHAT IT'S ACTUALLY DOING
                            # If ModulatoryProjection's value is not specified, try to assign one
                            if mod_proj_value is None:
                                # If not specified for Port, assign that
                                if VALUE not in port_dict or port_dict[VALUE] is None:
                                    port_dict[VALUE] = mod_signal_value
                                # If value has been assigned, make sure value is the same for ModulatorySignal
                                elif port_dict[VALUE] != mod_signal_value:
                                    # If the values differ, assign None so that Port's default is used
                                    port_dict[VALUE] = None
                                    # No need to check any more ModulatoryProjections
                                    break

                            #
                            else:
                                port_dict[VALUE] = mod_proj_value

                except ParameterPortError:
                    raise ParameterPortError("Tuple specification in {} specification dictionary "
                                          "for {} ({}) is not a recognized specification for one or more "
                                          "{}s, {}s, or {}s that project to it".
                                          format(ParameterPort.__name__,
                                                 owner.name,
                                                 projections_spec,
                                                 Mechanism.__name__,
                                                 ModulatorySignal.__name__,
                                                 Projection.__name__))

        elif port_specific_spec is not None:
            raise ParameterPortError("PROGRAM ERROR: Expected tuple or dict for {}-specific params but, got: {}".
                                  format(self.__class__.__name__, port_specific_spec))

        return port_spec, params_dict

    @staticmethod
    def _get_port_function_value(owner, function, variable):
        """Return parameter variable (since ParameterPort's function never changes the form of its variable"""
        return variable

    def _get_variable_from_projections(self, context=None):
        """
        Get backingfield ("base") value of param of function of Mechanism to which the ParameterPort belongs.
        """
        # FIX 3/6/19: source does not yet seem to have been assigned to owner.function
        return self.source._get(context)

    def get_label(self, context=None):
        raise ParameterPortError(f"{ParameterPort.__name__}s do not have labels.")

    @property
    def pathway_projections(self):
        raise ParameterPortError("PROGRAM ERROR: Attempt to access {} for {}; {}s do not have {}s".
                                  format(PATHWAY_PROJECTION, self.name, PARAMETER_PORT, PATHWAY_PROJECTION))

    @pathway_projections.setter
    def pathway_projections(self, value):
        raise ParameterPortError("PROGRAM ERROR: Attempt to assign {} to {}; {}s cannot accept {}s".
                                  format(PATHWAY_PROJECTION, self.name, PARAMETER_PORT, PATHWAY_PROJECTION))


def _instantiate_parameter_ports(owner, function=None, context=None):
    """Call _instantiate_parameter_port for all modulable parameters to instantiate ParameterPorts for them

    If owner.parameter_port is None or False:
        - no ParameterPorts will be instantiated.
    Otherwise, instantiate ParameterPort for each modulable parameter
    :param function:

    """

    # TBI / IMPLEMENT: use specs to implement ParameterPorts below

    owner._parameter_ports = ParameterPortList(
        component_type=ParameterPort,
        name=owner.name + '.parameter_ports',
        owner=owner,
    )

    # Check that all ParameterPorts for owner have not been explicitly suppressed
    try:
        if owner.parameter_ports is NotImplemented:
            return
    except KeyError:
        # PARAMETER_PORTS not specified at all, so OK to continue and construct them
        pass

    # Instantiate ParameterPort for each modulable Parameter on
    # function and owner. function is first because in some
    # cases a Parameter will be specified on both, and the function's
    # values/defaults should take precedence
    def skip_parameter_port(parameter):
        return (
            isinstance(parameter, (ParameterAlias, SharedParameter))
            or parameter.name in owner.exclude_from_parameter_ports
            or not parameter.modulable
        )

    def _enumerate_parameter_ports(obj, prev_objs, port_collection):
        """
        Returns:
            [Dict[str:List[List]]]: a dictionary containing keys that
            correspond to parameter names that will get parameter ports,
            and values of lists of lists. Each list contains a series of
            parameter names that correspond to a path to the function
            whose parameter (with name equivalent to the key) will be
            modulated
        """
        for p in obj.parameters:
            # prefer instantiated value if exists
            try:
                func = p._get(context)
            except FunctionError:
                func = None

            if func is None:
                func = p.default_value

            if (
                not p.reference
                and is_instance_or_subclass(func, Function)
                and not isinstance(p, (ParameterAlias, SharedParameter))
            ):
                _enumerate_parameter_ports(
                    func, prev_objs + [p.name], port_collection
                )

            if isinstance(p, ParameterAlias):
                port_aliases.add(p.name)

            if not skip_parameter_port(p):
                port_collection[p.name].append(prev_objs)

        return port_collection

    def _get_corresponding_component(obj, param_list, context=None, index=None):
        if index is None:
            index = len(param_list) - 1

        res = obj
        for p in param_list[0:index + 1]:
            param = getattr(res.parameters, p)
            func = param._get(context)

            if func is None:
                func = param.default_value

            res = func

        return res

    port_aliases = set()
    port_collection = _enumerate_parameter_ports(
        owner, [], collections.defaultdict(list)
    )

    for parameter_port_name in port_collection:
        add_suffix = (
            len(port_collection[parameter_port_name]) > 1
            or parameter_port_name in port_aliases
        )

        for corresponding_parameter_component_names in port_collection[parameter_port_name]:
            corresponding_parameter_component = _get_corresponding_component(
                owner, corresponding_parameter_component_names, context
            )

            p = getattr(
                corresponding_parameter_component.parameters,
                parameter_port_name
            )

            # .function is not finalized yet, because this happens before
            # _instantiate_function
            if corresponding_parameter_component is owner and p.name != FUNCTION:
                source = p
            else:
                source = operator.attrgetter(
                    f'{".".join(corresponding_parameter_component_names)}.parameters.{p.name}'
                )

            # use Shared/FunctionParameter value as fallback
            try:
                # want the isp value for the last parameter name of the
                # second to last component
                isp_owner = _get_corresponding_component(
                    owner, corresponding_parameter_component_names, context, -2
                )
            except IndexError:
                isp_owner = owner

            try:
                value = isp_owner.initial_shared_parameters[corresponding_parameter_component_names[-1]][p.name]
            except (AttributeError, IndexError, KeyError, TypeError):
                value = None

            # if parameter value on actual Parameter was specified or there is
            # no Shared/FunctionParameter value, use the actual Parameter default
            if p._user_specified or value is None:
                if p.spec is not None:
                    value = p.spec
                else:
                    value = p.default_value

            if add_suffix:
                explicit_name = ParameterPortList._get_explicit_name(
                    p.name,
                    corresponding_parameter_component_names
                    if len(corresponding_parameter_component_names) > 0
                    else ParameterPortList._owner_port_suffix
                )
            else:
                explicit_name = p.name

            _instantiate_parameter_port(
                owner,
                p.name,
                value,
                context=context,
                function=corresponding_parameter_component,
                source=source,
                explicit_name=explicit_name
            )

    owner.parameter_ports.sort(key=lambda port: port.name)

def _instantiate_parameter_port(
    owner,
    param_name,
    param_value,
    context,
    function=None,
    source=None,
    explicit_name=None
):
    """Call _instantiate_port for allowable params, to instantiate a ParameterPort for it

    Include ones in function.parameters
    Exclude if it is a:
        ParameterPort that already exists
        non-numeric value (including NotImplemented, False or True)
            unless it is:
                a tuple (could be one specifying Modulatory Component)
                a dict with the name FUNCTION_PARAMS (otherwise exclude)
        function or method
            IMPLEMENTATION NOTE: FUNCTION_RUNTIME_PARAM_NOT_SUPPORTED
            (this is because self.defaults.function could be a class rather than an bound method;
            i.e., not yet instantiated;  could be rectified by assignment in _instantiate_function)
    # FIX: UPDATE WITH MODULATION_MODS
    # FIX:    CHANGE TO IntegratorFunction FUnction ONCE LearningProjection MODULATES ParameterPort Function:
    If param_name is FUNCTION_PARAMS and param is a matrix (presumably for a MappingProjection)
        modify ParameterPort's function to be LinearCombination (rather Linear which is the default)
    """
    from psyneulink.core.components.ports.modulatorysignals.modulatorysignal import _is_modulatory_spec
    from psyneulink.core.components.projections.modulatory.modulatoryprojection import ModulatoryProjection_Base

    def _get_tuple_for_single_item_modulatory_spec(obj, name, value):
        """Return (<default param value>, <modulatory spec>) for modulatory spec
        """
        try:
            param_default_value = getattr(obj.defaults, name)
            # Only assign default value if it is not None
            if param_default_value is not None:
                return (param_default_value, value)
            else:
                return value
        except AttributeError:
            raise ParameterPortError("Unrecognized specification for {} paramater of {} ({})".
                                      format(param_name, owner.name, param_value))

    if explicit_name is None:
        explicit_name = param_name

    # EXCLUSIONS:

    # # Skip if ParameterPort already exists
    # if param_name in owner.ParameterPorts:
    #     return

    if param_value is NotImplemented:
        return
    # Allow numerics but omit booleans (which are treated by is_numeric as numerical)
    if is_numeric(param_value) and not isinstance(param_value, bool):
        pass
    # Only allow a FUNCTION_PARAMS dict
    elif isinstance(param_value, (ReadOnlyOrderedDict, dict)) and param_name == FUNCTION_PARAMS:
        pass
    # Allow ModulatoryProjection
    elif isinstance(param_value, Projection):
        if isinstance(param_value, ModulatoryProjection_Base):
            pass
        else:
            return
    # Allow Projection class
    elif inspect.isclass(param_value) and issubclass(param_value, Projection):
        if issubclass(param_value, (ModulatoryProjection_Base)):
            pass
        else:
            return

    elif _is_modulatory_spec(param_value, include_matrix_spec=False) and not isinstance(param_value, tuple):
        # If parameter is a single Modulatory specification (e.g., ControlSignal, or CONTROL, etc.)
        #   try to place it in a tuple (for interpretation by _parse_port_spec) using default value as 1st item
        #   (note: exclude matrix since it is allowed as a value specification but not a projection reference)
        try:
            param_value = _get_tuple_for_single_item_modulatory_spec(function, param_name, param_value)
        except ParameterPortError:
            param_value = _get_tuple_for_single_item_modulatory_spec(owner, param_name, param_value)

    # Allow tuples (could be spec that includes a Projection or Modulation)
    elif isinstance(param_value, tuple):
        # # FIX: EXTRACT VALUE HERE (AS IN Component.__init__?? [4/18/17]
        # param_value = owner._get_param_value_from_tuple(param_value)
        pass
    # Allow if it is a keyword for a parameter
    elif isinstance(param_value, str) and param_value in parameter_keywords:
        pass
    # Exclude function (see docstring above)
    elif param_name == FUNCTION:
        return
    # (7/19/17 CW) added this if statement below while adding `hetero` and `auto` and AutoAssociativeProjections: this
    # allows `hetero` to be specified as a matrix, while still generating a ParameterPort
    elif isinstance(param_value, np.ndarray):
        pass
    # allow function parameters
    elif param_name in function.parameters.names():
        pass
    # Exclude all others
    else:
        return

    # Assign ParameterPorts to Component for parameters of its function (function_params), except for ones that are:
    #    - another component
    #    - a function or method
    #    - have a value of None (see IMPLEMENTATION_NOTE below)
    #    - they have the same name as another parameter of the component (raise exception for this)

    # IMPLEMENTATION NOTE:
    # The following is necessary since, if ANY parameters of a function are specified, entries are made
    #    in the FUNCTION_PARAMS dict of its owner for ALL of the function's params;  however, their values
    #    will be set to None (and there may not be a way to determine a
    #    default; e.g., the length of the array for the weights or exponents params for LinearCombination).
    #    Therefore, None will be passed as the reference_value, which will cause validation of the
    #    ParameterPort's function (in _instantiate_function()) to fail.
    #  Current solution is to simply not instantiate a ParameterPort for any function_param that has
    #    not been explicitly specified
    if param_value is None:
        return

    if not _is_legal_param_value(owner, param_value):
        return

    elif (_is_modulatory_spec(param_value, include_matrix_spec=False)
            and not isinstance(param_value, tuple)):
        # If parameter is a single Modulatory specification (e.g., ControlSignal, or CONTROL, etc.)
        # try to place it in a tuple (for interpretation by _parse_port_spec) using default value as 1st item
        #   (note: exclude matrix since it is allowed as a value specification vs. a projection reference)
        try:
            param_value = _get_tuple_for_single_item_modulatory_spec(
                function,
                param_name,
                param_value
            )
        except ParameterPortError:
            param_value = _get_tuple_for_single_item_modulatory_spec(
                owner,
                param_name,
                param_value
            )

    # # FIX: 10/3/17 - ??MOVE THIS TO _parse_port_specific_specs ----------------
    # # Use param_value as constraint
    # # IMPLEMENTATION NOTE:  need to copy, since _instantiate_port() calls _parse_port_value()
    # #                       for constraints before port_spec, which moves items to subdictionaries,
    # #                       which would make them inaccessible to the subsequent parse of port_spec
    from psyneulink.core.components.ports.modulatorysignals.modulatorysignal import ModulatorySignal
    from psyneulink.core.components.mechanisms.modulatory.modulatorymechanism import ModulatoryMechanism_Base
    if (
        is_iterable(param_value)
        and any(isinstance(item, (ModulatorySignal, ModulatoryProjection_Base, ModulatoryMechanism_Base)) for item in param_value)
    ):
        reference_value = param_value
    else:
        reference_value = deepcopy(param_value)

    # Assign parameterPort for function_param to the component
    port = _instantiate_port(
        owner=owner,
        port_type=ParameterPort,
        name=explicit_name,
        port_spec=param_value,
        reference_value=reference_value,
        reference_value_name=param_name,
        params=None,
        context=context
    )
    if port:
        owner._parameter_ports[explicit_name] = port
        # will be parsed on assignment of function
        # FIX: if the function is manually changed after assignment,
        # FIX: the source will remain pointing to the original Function
        port.source = source
        # if the source parameter is not added here, we can't reference
        # a ParameterPort by Parameter
        owner.parameter_ports.parameter_mapping[source] = port

    return port


def _is_legal_param_value(owner, value):

    from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import _is_control_spec
    from psyneulink.core.components.mechanisms.modulatory.control.gating.gatingmechanism import _is_gating_spec

    # LEGAL PARAMETER VALUES:

    # # lists, arrays or numeric values
    if is_value_spec(value):
        return True

    # tuple, first item of which is a legal parameter value
    #     note: this excludes (param_name, Mechanism) tuples used to specify a ParameterPort
    #           (e.g., if specified for the control_signals param of ControlMechanism)
    if isinstance(value, tuple):
        if _is_legal_param_value(owner, value[0]):
            return True

    if isinstance(value, dict) and VALUE in value:
        return True

    if _is_control_spec(value) or _is_gating_spec(value):
        return True

    # keyword that resolves to one of the above
    if get_param_value_for_keyword(owner, value) is not None:
        return True

    # Assignment of ParameterPort for Component objects, function or method are not currently supported
    if isinstance(value, (types.FunctionType, types.MethodType, Component)):
        return False


def _get_parameter_port(sender_owner, sender_type, param_name, component):
    """Return ParameterPort for named parameter of a Mechanism requested by owner
    """

    # Validate that component is a Mechanism or Projection
    if not isinstance(component, (Mechanism, Projection)):
        raise ParameterPortError("Request for {} of a component ({}) that is not a {} or {}".
                                  format(PARAMETER_PORT, component, MECHANISM, PROJECTION))

    try:
        return component._parameter_ports[param_name]
    except KeyError:
        # Check that param (named by str) is an attribute of the Mechanism
        if not (hasattr(component, param_name) or hasattr(component.function, param_name)):
            raise ParameterPortError("{} (in specification of {}  {}) is not an attribute "
                                        "of {} or its function"
                                        .format(param_name, sender_type, sender_owner.name, component))
        # Check that the Mechanism has a ParameterPort for the param
        if param_name not in component._parameter_ports.names:
            raise ParameterPortError("There is no ParameterPort for the parameter ({}) of {} "
                                        "specified in {} for {}".
                                        format(param_name, component.name, sender_type, sender_owner.name))



import timeit

import numpy as np
from psyneulink import *

from gym_forager.envs.forager_env import ForagerEnv

# Runtime Switches:
PNL=True
RENDER = False
PERCEPT_DISTORT = False
PNL_COMPILE = True

# *********************************************************************************************************************
# *********************************************** CONSTANTS ***********************************************************
# *********************************************************************************************************************

# These should probably be replaced by reference to ForagerEnv constants:
obs_len = 2
action_len = 2
player_coord_idx = slice(0,2)
predator_coord_idx = slice(3,5)
prey_coord_idx = slice(6,8)
player_value_idx = 2
predator_value_idx = 5
prey_value_idx = 8

player_len = prey_len = predator_len = obs_len

# *********************************************************************************************************************
# **************************************  MECHANISMS AND COMPOSITION  *************************************************
# *********************************************************************************************************************

if PERCEPT_DISTORT:
    player = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort(variance=0), name="PLAYER OBS")
    prey = ProcessingMechanism(input_shapes=prey_len, function=GaussianDistort(variance=0), name="PREY OBS")
else:
    player = TransferMechanism(input_shapes=prey_len, name="PLAYER OBS")
    prey = TransferMechanism(input_shapes=prey_len, name="PREY OBS")

# For future use:
values = TransferMechanism(input_shapes=3, name="AGENT VALUES")
reward = TransferMechanism(name="REWARD")

# Use ComparatorMechanism to compute direction of action as difference of coordinates between player and prey:
# note: unitization is done in main loop, to allow compilation of LinearCombination function) (TBI)
greedy_action_mech = ComparatorMechanism(name='MOTOR OUTPUT',sample=player,target=prey)

agent_comp = Composition(name='PREDATOR-PREY COMPOSITION')

agent_comp.add_node(player)
agent_comp.add_node(prey)
agent_comp.add_node(greedy_action_mech)

# agent_comp.show_graph()


# *********************************************************************************************************************
# ******************************************   RUN SIMULATION  ********************************************************
# *********************************************************************************************************************

num_trials = 100

def main():
    env = ForagerEnv()
    reward = 0
    done = False
    if RENDER:
        env.render()  # If visualization is desired
    else:
        print("Running simulation...")
    steps = 0
    start_time = timeit.default_timer()
    for _ in range(num_trials):
        observation = env.reset()
        while True:
            if PNL:
                if PNL_COMPILE:
                    BIN_EXECUTE = 'LLVMRun'
                else:
                    BIN_EXECUTE = 'Python'
                run_results = agent_comp.run(inputs={player:[observation[player_coord_idx]],
                                                     prey:[observation[prey_coord_idx]],
                                                     },
                                             bin_execute=BIN_EXECUTE
                                             )
                run_results[0] = np.array(run_results[0])
                action= np.where(run_results[0] == 0, 0, run_results[0] / np.abs(run_results[0]))
            else:
                run_results = observation[prey_coord_idx] - observation[player_coord_idx]
                action= np.where(run_results == 0, 0, run_results / np.abs(run_results))

            observation, reward, done, _ = env.step(action)
            steps +=1
            if done:
                break
    stop_time = timeit.default_timer()
    print(f'{steps / (stop_time - start_time):.1f} steps/second, {steps} total steps in '
          f'{stop_time - start_time:.2f} seconds')
    if RENDER:
        env.render()  # If visualization is desired

if __name__ == "__main__":
    main()



from . import basepreferenceset
from . import compositionpreferenceset
from . import mechanismpreferenceset
from . import preferenceset

from .basepreferenceset import *
from .compositionpreferenceset import *
from .mechanismpreferenceset import *
from .preferenceset import *

__all__ = list(basepreferenceset.__all__)
__all__.extend(compositionpreferenceset.__all__)
__all__.extend(mechanismpreferenceset.__all__)
__all__.extend(preferenceset.__all__)



import numpy as np
import pytest
import random

from psyneulink.core import llvm as pnlvm

SEED = 0

@pytest.mark.benchmark(group="Mersenne Twister bounded integer PRNG")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize("bounds, expected",
    [((0xffffffff,), [3626764237, 1654615998, 3255389356, 3823568514, 1806341205]),
     ((14,), [13, 12,  2,  5,  4]),
     ((0,14), [13, 12,  2,  5,  4]),
     ((5,0xffff), [2002, 28611, 19633,  1671, 37978]),
    ], ids=lambda x: str(x) if len(x) != 5 else "")
# Python uses sampling of upper bits (vs. lower bits in Numpy). Skip it in this test.
def test_random_int32_bounded(benchmark, mode, bounds, expected):

    if mode == 'numpy':
        # Numpy promotes elements to int64
        state = np.random.RandomState([SEED])

        def f():
            return state.randint(*bounds, dtype=np.uint32)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state = init_fun.np_buffer_for_arg(0)

        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_int32_bounded')

        def f():
            lower, upper = bounds if len(bounds) == 2 else (0, bounds[0])
            out = gen_fun.np_buffer_for_arg(3)
            gen_fun(state, lower, upper, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')

        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)

        init_fun.cuda_call(gpu_state, np.int32(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_int32_bounded')
        out = gen_fun.np_buffer_for_arg(3)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            lower, upper = bounds if len(bounds) == 2 else (0, bounds[0])
            gen_fun.cuda_call(gpu_state, np.uint32(lower), np.uint32(upper), gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f(), f(), f(), f(), f()]
    np.testing.assert_array_equal(res, expected)
    benchmark(f)

@pytest.mark.benchmark(group="Mersenne Twister integer PRNG")
@pytest.mark.parametrize('mode', ['Python', 'numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
def test_random_int32(benchmark, mode):

    if mode == 'Python':
        state = random.Random(SEED)

        def f():
            return state.randrange(0xffffffff)

    elif mode == 'numpy':
        # Numpy promotes elements to int64
        state = np.random.RandomState([SEED])

        def f():
            return state.randint(0xffffffff, dtype=np.uint32)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state = init_fun.np_buffer_for_arg(0)

        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_int32')

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')

        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)

        init_fun.cuda_call(gpu_state, np.int32(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_int32')
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f(), f(), f(), f(), f()]
    np.testing.assert_array_equal(res, [3626764237, 1654615998, 3255389356, 3823568514, 1806341205])
    benchmark(f)


@pytest.mark.benchmark(group="Mersenne Twister floating point PRNG")
@pytest.mark.parametrize('mode', ['Python', 'numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
def test_random_float(benchmark, mode):

    if mode == 'Python':
        # Python treats every seed as array
        state = random.Random(SEED)

        def f():
            return state.random()

    elif mode == 'numpy':
        # numpy promotes elements to int64
        state = np.random.RandomState([SEED])

        def f():
            return state.random_sample()

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_double')

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')

        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)

        init_fun.cuda_call(gpu_state, np.int32(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_double')
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f(), f()]
    np.testing.assert_allclose(res, [0.8444218515250481, 0.7579544029403025])
    benchmark(f)


@pytest.mark.benchmark(group="Marsenne Twister Normal distribution")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
# Python uses different algorithm so skip it in this test
def test_random_normal(benchmark, mode):

    if mode == 'numpy':
        # numpy promotes elements to int64
        state = np.random.RandomState([SEED])

        def f():
            return state.normal()

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_normal')

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')

        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)

        init_fun.cuda_call(gpu_state, np.int32(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_normal')
        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f(), f()]
    np.testing.assert_allclose(res, [0.4644982638709743, 0.6202001216069017], rtol=1e-5, atol=1e-8)
    benchmark(f)

@pytest.mark.benchmark(group="Marsenne Twister Binomial distribution")
@pytest.mark.parametrize('mode', ['numpy',
                                  pytest.param('LLVM', marks=pytest.mark.llvm),
                                  pytest.helpers.cuda_param('PTX')])
@pytest.mark.parametrize('n', [1])
@pytest.mark.parametrize('p, exp', [
    (0, [0]),
    (0.1, [0x20d00c]),
    (0.33, [0xc224f70d]),
    (0.5, [0xca76f71d]),
    (0.66, [0x3ddb08f2]),
    (0.95, [0xffffbffb]),
    (1, [0xffffffff]),
    ])
# Python uses different algorithm so skip it in this test
def test_random_binomial(benchmark, mode, n, p, exp):
    if mode == 'numpy':
        # numpy promotes elements to int64
        state = np.random.RandomState([SEED])

        def f():
            return state.binomial(n, p)

    elif mode == 'LLVM':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')
        state = init_fun.np_buffer_for_arg(0)
        init_fun(state, SEED)

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_binomial')
        n = np.asarray(n, dtype=gen_fun.np_arg_dtypes[1])
        p = np.asarray(p, dtype=gen_fun.np_arg_dtypes[2])

        def f():
            out = gen_fun.np_buffer_for_arg(1)
            gen_fun(state, n, p, out)
            return out

    elif mode == 'PTX':
        init_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_init')

        state_size = init_fun.np_buffer_for_arg(0).nbytes
        gpu_state = pnlvm.jit_engine.pycuda.driver.mem_alloc(state_size)

        init_fun.cuda_call(gpu_state, np.int32(SEED))

        gen_fun = pnlvm.LLVMBinaryFunction.get('__pnl_builtin_mt_rand_binomial')

        gpu_n = pnlvm.jit_engine.pycuda.driver.In(np.asarray(n, dtype=gen_fun.np_arg_dtypes[1]))
        gpu_p = pnlvm.jit_engine.pycuda.driver.In(np.asarray(p, dtype=gen_fun.np_arg_dtypes[2]))

        out = gen_fun.np_buffer_for_arg(1)
        gpu_out = pnlvm.jit_engine.pycuda.driver.Out(out)

        def f():
            gen_fun.cuda_call(gpu_state, gpu_n, gpu_p, gpu_out)
            return out.copy()

    else:
        assert False, "Unknown mode: {}".format(mode)

    res = [f() for _ in range(32)]
    res = int(''.join(str(x) for x in res), 2)
    assert res == exp[n - 1]
    benchmark(f)



import gc
import numpy as np
import pytest
import weakref

import graph_scheduler as gs
import psyneulink as pnl

show_graph_args = ["show_all", "show_node_structure", "show_cim", "show_learning", "show_types", "show_dimensions",
                   "show_projection_labels", "show_projections_not_in_composition"]

@pytest.mark.composition
@pytest.mark.parametrize("show_graph_args", [pytest.param(None, id="show_graph_disabled"),
                                             pytest.param({}, id="show_graph_default"),
                                             *(pytest.param({arg: True}, id=arg) for arg in show_graph_args),
                                            ])
@pytest.mark.parametrize("op", ["construct", "run"])
def test_composition_leak(comp_mode, op, show_graph_args):

    c = pnl.Composition()
    t = pnl.TransferMechanism()
    c.add_node(t)

    if op == "run":
        res = c.run([5], execution_mode=comp_mode)
        np.testing.assert_array_equal(res, [[5]])

    if show_graph_args is not None:
        c.show_graph(**show_graph_args, output_fmt=None)

    weak_c = weakref.ref(c)
    weak_t = weakref.ref(t)

    # Clear all known global references
    for registry in pnl.primary_registries:
        pnl.clear_registry(registry)

    pnl.core.llvm.LLVMBinaryFunction.get.cache_clear()
    pnl.core.llvm.LLVMBinaryFunction.from_obj.cache_clear()

    gs.utilities.cached_hashable_graph_function.cache_clear()

    # Remove the original references
    del t
    del c

    gc.collect()

    assert weak_c() is None, gc.get_referrers(weak_c())
    assert weak_t() is None, gc.get_referrers(weak_t())



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ***************************************************  DDM *************************************************************

"""

Contents
--------
  * `DDM_Overview`
  * `DDM_Creation`
  * `DDM_Structure`
      - `DDM_Input`
      - `DDM_Output`
          • `Default OutputPorts <DDM_Default_OutputPorts>`
          • `Custom OutputPorts <DDM_Custom_OutputPorts>`
      - `DDM_Modes`
          • `DDM_Analytic_Mode`
          • `DDM_Integration_Mode`
  * `DDM_Execution`
  * `DDM_Class_Reference`


.. _DDM_Overview:

Overview
--------
The DDM Mechanism implements the "Drift Diffusion Model" (also know as the Diffusion Decision, Accumulation to Bound,
Linear IntegratorFunction, and `First Passage Time Model <https://en.wikipedia.org/wiki/First-hitting-time_model>`_
for a `Wiener Process <https://en.wikipedia.org/wiki/Wiener_process>`_. This corresponds to a continuous version
of the `sequential probability ratio test (SPRT) <https://en.wikipedia.org/wiki/Sequential_probability_ratio_test>`_,
that is the statistically optimal procedure for `two alternative forced choice (TAFC) decision making
<https://en.wikipedia.org/wiki/Two-alternative_forced_choice>`_ (see `drift-diffusion model
<https://en.wikipedia.org/wiki/Two-alternative_forced_choice#Drift-diffusion_model>`_ in partciular).

The DDM Mechanism may be constructed with a choice of several functions that fall into to general categories: analytic
solutions and path integration (see `DDM_Modes` below for more about these options.)

.. _DDM_Creation:

Creating a DDM Mechanism
-----------------------------
A DDM Mechanism can be instantiated directly by calling its constructor, or by using the `mechanism` command and
specifying DDM as its **mech_spec** argument.  The model implementation is selected using the `function <DDM.function>`
argument. The function selection can be simply the name of a DDM function::

    >>> import psyneulink as pnl
    >>> my_DDM = pnl.DDM(function=pnl.DriftDiffusionAnalytical)

or a call to the function with arguments specifying its parameters::

    >>> my_DDM = pnl.DDM(function=pnl.DriftDiffusionAnalytical(drift_rate=0.2, threshold=1.0))


COMMENT:
.. _DDM_Input:
**Input**.  The `default_variable` argument specifies the default value to use as the stimulus component of the
`drift rate <DDM_Drift_Rate>` for the decision process.  It must be a single scalar value.
[TBI - MULTIPROCESS DDM - REPLACE ABOVE]
**Input**.  The ``default_variable`` argument specifies the default value to use as the stimulus component of the
`drift rate <DDM_Drift_Rate>` for each decision process, as well as the number of decision processes implemented
and the corresponding format of the ``input`` required by calls to its ``execute`` and ``run`` methods.  This can be a
single scalar value or an an array (list or 1d np.array). If it is a single value (as in the first two examples above),
a single DDM process is implemented, and the input is used as the stimulus component of the
`drift rate <DDM_Drift_Rate>` for the process. If the input is an array (as in the third example above),
multiple parallel DDM processes are implemented all of which use the same parameters but each of which receives its
own input (from the corresponding element of the input array) and is executed independently of the others.
COMMENT

.. _DDM_Structure:

Structure
---------

The DDM Mechanism implements a general form of the decision process.

.. _DDM_Input:

*Input*
~~~~~~~

The input to the `function <DDM_Function>` of a DDM Mechanism is always a scalar, irrespective of `type of function
<DDM_Modes>` that is used.  Accordingly, the default `InputPort` for a DDM takes a single scalar value as its input,
that represents the stimulus for the decision process.  However, this can be configured using the **input_format**
argument of the DDM's consructor, to accomodate use of the DDM with other Mechanisms that generate a stimulus array
(e.g., representing the stimuli associated with each of the two choices). By default, the **input_format** is
*SCALAR*.  However, if it is specified as *ARRAY*, the DDM's InputPort is configured to accept a 1d 2-item vector,
and to use `Reduce` as its Function, which subtracts the 2nd element of the vector from the 1st, and provides this as
the input to the DDM's `function <DDM.function>`.  If *ARRAY* is specified, two  `Standard OutputPorts
<OutputPorts_Standard>` are added to the DDM, that allow the result of the decision process to be represented
as an array corresponding to the input array (see `below <DDM_Custom_OutputPorts>`).

COMMENT:
ADD EXAMPLE HERE
COMMENT

COMMENT
NOTE SURE WHAT THIS MEANS:
That parameter, along with all of the others for the DDM, must be assigned as
parameters of the DDM's `function <DDM.function>` (see examples under `DDM_Modes` below, and individual `Functions
<Function>` for additional details).
COMMENT

.. _DDM_Output:

*Output*
~~~~~~~~

The DDM Mechanism can generate two different types of results depending on which function is selected. When a
function representing an analytic solution is selected, the mechanism generates a single estimation for the process.
When the path integration function is selected, the mechanism carries out step-wise integration of the process; each
execution of the mechanism computes one step. (see `DDM_Modes` and `DDM_Execution` for additional details).

The `value <DDM.value>` of the DDM Mechanism may have up to ten items. The first two of these are always assigned, and
are represented by the DDM Mechanism's two default `OutputPorts``: `DECISION_VARIABLE <DDM_DECISION_VARIABLE>` and
`RESPONSE_TIME <DDM_RESPONSE_TIME>`.  Other `output_ports <DDM.output_ports>` may be automatically assigned,
depending on the `function <DDM.function>` that has been assigned to the DDM, as shown in the table below:

.. _DDM_Default_OutputPorts:

+------------------------------------+---------------------------------------------------------+
|                                    |                     **Function**                        |
|                                    |                      *(type)*                           |
+                                    +----------------------------+----------------------------+
|                                    | `DriftDiffusionAnalytical` | `DriftDiffusionIntegrator` |
|                                    |   (`analytic               |   (`path integration)      |
| **OutputPorts:**                   |   <DDM_Analytic_Mode>`)    |   <DDM_Integration_Mode>`) |
+------------------------------------+----------------------------+----------------------------+
| `DECISION_VARIABLE                 |                            |                            |
| <DDM_DECISION_VARIABLE>`           |       X                    |          X                 |
+------------------------------------+----------------------------+----------------------------+
| `RESPONSE_TIME                     |                            |                            |
| <DDM_RESPONSE_TIME>`               |       X                    |          X                 |
+------------------------------------+----------------------------+----------------------------+
| `PROBABILITY_UPPER_THRESHOLD       |                            |                            |
| <DDM_PROBABILITY_UPPER_THRESHOLD>` |       X                    |                            |
+------------------------------------+----------------------------+----------------------------+
| `PROBABILITY_LOWER_THRESHOLD       |                            |                            |
| <DDM_PROBABILITY_LOWER_THRESHOLD>` |       X                    |                            |
+------------------------------------+----------------------------+----------------------------+
| `RT_CORRECT_MEAN                   |                            |                            |
| <DDM_RT_CORRECT_MEAN>`             |       X                    |                            |
+------------------------------------+----------------------------+----------------------------+
| `RT_CORRECT_VARIANCE               |                            |                            |
| <DDM_RT_CORRECT_VARIANCE>`         |       X                    |                            |
+------------------------------------+----------------------------+----------------------------+
| `RT_CORRECT_SKEW                   |                            |                            |
| <DDM_RT_CORRECT_SKEW>`             |       X                    |                            |
+------------------------------------+----------------------------+----------------------------+
| `RT_INCORRECT_MEAN                 |                            |                            |
| <DDM_RT_INCORRECT_MEAN>`           |       X                    |                            |
+------------------------------------+----------------------------+----------------------------+
| `RT_INCORRECT_VARIANCE             |                            |                            |
| <DDM_RT_INCORRECT_VARIANCE>`       |       X                    |                            |
+------------------------------------+----------------------------+----------------------------+
| `RT_INCORRECT_SKEW                 |                            |                            |
| <DDM_RT_INCORRECT_SKEW>`           |       X                    |                            |
+------------------------------------+----------------------------+----------------------------+

.. _DDM_Custom_OutputPorts:

The `output_ports <DDM.output_ports>` assigned to a DDM can explicilty determined by specifying ones from its list
of `standard_output_ports <DDM.standard_output_ports>` in the **output_ports** argument of its constructor, or the
*OUTPUT_PORTS* entry of an `OutputPort specification dictionary <OutputPort_Specification_Dictionary>`.  This
can include any of the OutputPorts `listed above `DDM_Default_OutputPorts`, as well as two additional ones --
-- `DECISION_VARIABLE_ARRAY <DDM_DECISION_VARIABLE_ARRAY>` and `SELECTED_INPUT_ARRAY <DDM_SELECTED_INPUT_ARRAY>` --
that are available if the *ARRAY* option is specified in its **input_format** argument (see `DDM_Input`).  All of
these `Standard OutputPorts <OutputPort_Standard>` are listed in the DDM's `standard_output_ports
<DDM.standard_output_ports>` attribute. As with any Mechanism, `customized OutputPorts <OutputPort_Customization>`
can also be created and assigned.

.. _DDM_Modes:

*DDM Function Types*
~~~~~~~~~~~~~~~~~~~~

.. _DDM_Analytic_Mode:

Analytic Solutions
^^^^^^^^^^^^^^^^^^

The Drift Diffusion Model `Functions <Function>` that calculate analytic solutions
[Bogacz et al (2006), Srivastava et al. (2016)] is `DriftDiffusionAnalytical <DriftDiffusionAnalytical>`
`function <DDM.function>`, the mechanism generates a single estimate of the outcome for the decision process (see
`DDM_Execution` for details). In addition to `DECISION_VARIABLE <DDM_DECISION_VARIABLE>` and
`RESPONSE_TIME <DDM_RESPONSE_TIME>`, the Function returns an accuracy value (represented in the
`PROBABILITY_UPPER_THRESHOLD <DDM_PROBABILITY_UPPER_THRESHOLD>` OutputPort), and an error rate value (in the
`PROBABIILITY_LOWER_THRESHOLD <DDM_PROBABILITY_LOWER_THRESHOLD>` OutputPort, and moments (mean, variance, and skew)
for conditional (correct\\positive or incorrect\\negative) response time distributions. These are; the mean RT for
correct responses  (`RT_CORRECT_MEAN <DDM_RT_CORRECT_MEAN>`, the RT variance for correct responses
(`RT_CORRECT_VARIANCE <DDM_RT_CORRECT_VARIANCE>`, the RT skew for correct responses
(`RT_CORRECT_SKEW <DDM_RT_CORRECT_SKEW>`, the mean RT for incorrect responses  (`RT_INCORRECT_MEAN
<DDM_RT_INCORRECT_MEAN>`, the RT variance for incorrect responses (`RT_INCORRECT_VARIANCE
<DDM_RT_INCORRECT_VARIANCE>`, the RT skew for incorrect responses (`RT_INCORRECT_SKEW <DDM_RT_INCORRECT_SKEW>`.

An example that illustrate all of the parameters is shown below:

`DriftDiffusionAnalytical <DriftDiffusionAnalytical>` Function::

    >>> my_DDM_DriftDiffusionAnalytical = pnl.DDM(
    ...     function=pnl.DriftDiffusionAnalytical(
    ...         drift_rate=0.08928,
    ...         starting_value=0.5,
    ...         threshold=0.2645,
    ...         noise=0.5,
    ...         non_decision_time=0.15
    ...     ),
    ...     name='my_DDM_DriftDiffusionAnalytical'
    ... )

.. _DDM_Integration_Mode:

Path Integration
^^^^^^^^^^^^^^^^

The Drift Diffusion Model `Function <Function>` that calculates a path integration is `DriftDiffusionIntegrator
<DriftDiffusionIntegrator>`. The DDM Mechanism uses the `Euler method <https://en.wikipedia.org/wiki/Euler_method>`_ to
carry out numerical step-wise integration of the decision process (see `Execution <DDM_Execution>` below).  In this
mode, only the `DECISION_VARIABLE <DDM_DECISION_VARIABLE>` and `RESPONSE_TIME <DDM_RESPONSE_TIME>` are available.

`IntegratorFunction <IntegratorFunction>` Function::

    >>> my_DDM_path_integrator = pnl.DDM(
    ...     function=pnl.DriftDiffusionIntegrator(
    ...         noise=0.5,
    ...         initializer=1.0,
    ...         non_decision_time=2.0,
    ...         rate=3.0
    ...     ),
    ...     name='my_DDM_path_integrator'
    ... )

COMMENT:
[TBI - MULTIPROCESS DDM - REPLACE ABOVE]
The DDM Mechanism implements a general form of the decision process.  A DDM Mechanism assigns one **inputPort** to
each item in the `default_variable` argument, corresponding to each of the decision processes implemented
(see `Input <DDM_Input>` above). The decision process can be configured to execute in different modes.  The
`function <DDM.function>` parameters is the primary determinants of how the
decision process is executed, and what information is returned. The `function <DDM.function>` parameter specifies
the analytical solution to use. The number of `OutputPorts <OutputPort>` is determined by the `function <DDM.function>`
in use (see `list of output values <DDM_Results>` below).

[TBI - average_output_ports ARGUMENT/OPTION AFTER IMPLEMENTING MULTIPROCESS DDM]
OUTPUT MEASURE?? OUTCOME MEASURE?? RESULT?? TYPE OF RESULT??
If only a single decision process was run, then the value of each outputPort is the corresponding output of
the decision process.  If there is more than one decision process (i.e., the input has more than one item), then
the content of the outputPorts is determined by the ``average_output_ports`` argument.  If it is `True`,
then each outputPort (and item of ``output_values``) contains a single value, which is the average of the output
values of that type over all of the processes run.  If ``average_output_ports`` is :keyword:`False` (the default),
then the value of each ouputState is a 1d array, each element of which is the outcome of that type for the
corresponding decision process.
COMMENT


COMMENT:  [OLD;  PUT SOMEHWERE ELSE??]

    The DDM process uses the same set of parameters for all modes of execution.  These can be specified as arguments
    for the functions used in TRIAL mode, or in a params dictionary assigned to the `params` argument,
    using the keywords in the list below, as in the following example::
        my_DDM = DDM(
            function=DriftDiffusionAnalytical(drift_rate=0.1),
            params={
                DRIFT_RATE:(0.2, ControlProjection),
                STARTING_VALUE:-0.5
            },
        )
    The parameters for the DDM when `function <DDM.function>` is set to `DriftDiffusionAnalytical` are:

    .. _DDM_Drift_Rate:

    * `DRIFT_RATE <drift_rate>` (default 0.0)
      - multiplies the input to the Mechanism before assigning it to the `variable <DDM.variable>` on each call of
      `function <DDM.function>`.  The resulting value is further multiplied by the value of any ControlProjections to
      the `DRIFT_RATE` parameterPort. The `drift_rate` parameter can be thought of as the "automatic" component
      (baseline strength) of the decision process, the value received from a ControlProjection as the "attentional"
      component, and the input its "stimulus" component.  The product of all three determines the drift rate in
      effect for each time_step of the decision process.
    ..
    * `STARTING_VALUE <starting_value>` (default 0.0)
      - specifies the starting value of the decision variable.
    ..
    * `THRESHOLD` (default 1.0)
      - specifies the stopping value for the decision process. The `threshold` parameter must be greater than or
      equal to zero.
    ..
    * `NOISE` (default 0.5)
      - specifies the variance of the stochastic ("diffusion") component of the decision process.
    ..
    * `NON_DECISION_TIME` (default 0.2)
      specifies the `non_decision_time` parameter of the decision process (in units of seconds).

[TBI - MULTIPROCESS DDM - REPLACE BELOW]
When a DDM Mechanism is executed it computes the decision process, either analytically (in TRIAL mode)
or by step-wise integration (in TIME_STEP mode).  As noted above, if the input is a single value,
it computes a single DDM process.  If the input is a list or array, then multiple parallel DDM processes are executed,
with each element of the input used for the corresponding process.  All use the same set of parameters,
so the analytic solutions (used in TRIAL mode) for a given input will be the same; to implement processes in
this mode that use different parameters, a separate DDM Mechanism should explicitly be created for each. In
TIME_STEP mode, the noise term will resolve to different values in each time step, so the integration
paths and outcomes for the same input value will vary. This can be used to generate distributions of the process for a
single set of parameters that are not subject to the analytic solution (e.g., for time-varying drift rates).

.. note::
   DDM handles "runtime" parameters (specified in a call to its
   :py:meth:`execute <Mechanism_Base.execte>` or :py:meth:`run <Mechanism_Base.run>` methods)
   differently than standard Components: runtime parameters are added to the Mechanism's current value of the
   corresponding ParameterPort (rather than overriding it);  that is, they are combined additively with the value of
   any `ControlProjection` it receives to determine the parameter's value for that execution.  The ParameterPort's
   value is then restored to its original value (i.e., either its default value or the one assigned when it was
   created) for the next execution.

  ADD NOTE ABOUT INTERROGATION PROTOCOL, USING ``terminate_function``
  ADD NOTE ABOUT RELATIONSHIP OF RT TO time_steps TO non_decision_time TO ms
COMMENT

.. _DDM_Execution:

Execution
---------

When a DDM Mechanism is executed, it computes the decision process either `analytically <DDM_Analytic_Mode>`  or by
`numerical step-wise integration <DDM_Integration_Mode>` of its path.  The method used is determined by its `function
<DDM.function>` (see `DDM_Modes`). The DDM's `function <DDM.function>` always returns values for the `DECISION_VARIABLE
<DDM_DECISION_VARIABLE>` and `RESPONSE_TIME <DDM_RESPONSE_TIME>`, and assigns these as the first two items of its `value
<DDM.value>` attribute, irrespective of its function.

When an `analytic <DDM_Analytic_Mode>` function is selected, the same set of values is returned for every execution.
The returned values are determined entirely by the set of parameters passed to its `function <DDM.function>`.

When the `path integration <DDM_Integration_Mode>`, function is selected, a single step of integration is conducted each
time the Mechanism is executed. The returned values accumulate on every execution.

The analytic functions return a final positon and time of the model, along with other statistics, where as the path
integration function returns intermediate position and time values. The two types of functions can be thought of as
happening on different time scales: trial (analytic) and time step (path integration).

References
----------

*   Bogacz, R., Brown, E., Moehlis, J., Holmes, P., & Cohen, J. D. (2006).
    The physics of optimal decision making: A formal analysis of models of performance in two-alternative forced-choice
    tasks. Psychological Review, 113(4), 700-765.   http://dx.doi.org/10.1037/0033-295X.113.4.700

*   Srivastava, V., Holmes, P., Simen., P. (2016).
    Explicit moments of decision times for single- and double-threshold drift-diffusion processes,
    Journal of Mathematical Psychology, 75, 96-109,
    ISSN 0022-2496,
    https://doi.org/10.1016/j.jmp.2016.03.005.
    (http://www.sciencedirect.com/science/article/pii/S0022249616000341)

.. _DDM_Class_Reference:

Class Reference
---------------
"""
import sys
import logging
import types
from collections.abc import Iterable

import numpy as np
from beartype import beartype

from psyneulink._typing import Optional, Union, Literal

from psyneulink.core.components.functions.function import DEFAULT_SEED, _random_state_getter, _seed_setter
from psyneulink.core.components.functions.stateful.integratorfunctions import \
    DriftDiffusionIntegrator, IntegratorFunction
from psyneulink.core.components.functions.nonstateful.distributionfunctions import STARTING_VALUE, \
    DriftDiffusionAnalytical
from psyneulink.core.components.functions.nonstateful.transformfunctions import Reduce
from psyneulink.core.components.mechanisms.modulatory.control.controlmechanism import _is_control_spec
from psyneulink.core.components.mechanisms.mechanism import MechanismError
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.components.ports.modulatorysignals.controlsignal import ControlSignal
from psyneulink.core.components.ports.outputport import SEQUENTIAL, StandardOutputPorts
from psyneulink.core.globals.context import ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import \
    ALLOCATION_SAMPLES, FUNCTION, FUNCTION_PARAMS, INPUT_PORT_VARIABLES, NAME, OWNER_VALUE, \
    THRESHOLD, VARIABLE, PREFERENCE_SET_NAME
from psyneulink.core.globals.parameters import Parameter, check_user_specified
from psyneulink.core.globals.preferences.basepreferenceset import ValidPrefSet, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel
from psyneulink.core.globals.utilities import convert_all_elements_to_np_array, is_numeric, is_same_function_spec, object_has_single_value, get_global_seed
from psyneulink.core.scheduling.condition import AtTrialStart
from psyneulink.core.components.functions.userdefinedfunction import UserDefinedFunction

from psyneulink.core import llvm as pnlvm



__all__ = [
    'ARRAY', 'DDM', 'DDMError', 'DECISION_VARIABLE', 'DECISION_VARIABLE_ARRAY',
    'PROBABILITY_LOWER_THRESHOLD', 'PROBABILITY_UPPER_THRESHOLD', 'RESPONSE_TIME',
    'RT_CORRECT_MEAN', 'RT_CORRECT_SKEW', 'RT_CORRECT_VARIANCE',
    'RT_INCORRECT_MEAN', 'RT_INCORRECT_SKEW', 'RT_INCORRECT_VARIANCE',
    'SCALAR', 'SELECTED_INPUT_ARRAY', 'VECTOR', 'DECISION_OUTCOME'
]

logger = logging.getLogger(__name__)

DEFAULT_VARIABLE = 0.0

DECISION_VARIABLE = 'DECISION_VARIABLE'
DECISION_VARIABLE_ARRAY = 'DECISION_VARIABLE_ARRAY'
DECISION_OUTCOME = 'DECISION_OUTCOME'
SELECTED_INPUT_ARRAY = 'SELECTED_INPUT_ARRAY'
RESPONSE_TIME = 'RESPONSE_TIME'
PROBABILITY_UPPER_THRESHOLD = 'PROBABILITY_UPPER_THRESHOLD'
PROBABILITY_LOWER_THRESHOLD = 'PROBABILITY_LOWER_THRESHOLD'
RT_CORRECT_MEAN = 'RT_CORRECT_MEAN'
RT_CORRECT_VARIANCE = 'RT_CORRECT_VARIANCE'
RT_CORRECT_SKEW = 'RT_CORRECT_SKEW'
RT_INCORRECT_MEAN = 'RT_INCORRECT_MEAN'
RT_INCORRECT_VARIANCE = 'RT_INCORRECT_VARIANCE'
RT_INCORRECT_SKEW = 'RT_INCORRECT_SKEW'


# input_format Keywords:
SCALAR='SCALAR'
ARRAY='ARRAY'
VECTOR='VECTOR'

def decision_variable_to_array(x):
    """Generate "one-hot" 1d array designating selected action from DDM's scalar decision variable
    (used to generate value of OutputPort for action_selection Mechanism
    """
    if x >= 0:
        return [x,0]
    else:
        return [0,x]


class DDMError(MechanismError):
    pass


class DDM(ProcessingMechanism):
    # DOCUMENT:   COMBINE WITH INITIALIZATION WITH PARAMETERS
    #             ADD INFO ABOUT B VS. N&F
    #             ADD _instantiate_output_ports TO INSTANCE METHODS, AND EXPLAIN RE: NUM OUTPUT VALUES FOR B VS. N&F
    """
    DDM(                                   \
        default_variable=None,             \
        function=DriftDiffusionAnalytical)

    Implements a drift diffusion process (also known as the `Diffusion Decision Model
    <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2474742/>`_, either by calculating an `analytic solution
    <DDM_Analytic_Mode>` or carrying out `step-wise numerical integration <DDM_Integration_Mode>`.
    See `Mechanism <Mechanism_Class_Reference>` for additional arguments and attributes.

    The default behaviour for the DDM is to reset its integration state on each new trial, this can be overridden by
    setting the `reset_stateful_function_when <Component.reset_stateful_function_when>` attribute to a different
    condition. In addition, unlike `TransferMechanism <TransferMechanism>`, the DDM's `execute_until_finished
    <Component.execute_until_finished>` attribute is set to `True` by default. This will cause the DDM to execute
    multiple time steps per call to its `execute <Component.execute>`  method until the decision threshold is reached.
    This default behavior can be changed by setting `execute_until_finished <Component.execute_until_finished>` to
    `False`.


    Arguments
    ---------

    default_variable : value, list or np.ndarray : default FUNCTION_PARAMS[STARTING_VALUE]
        the input to the Mechanism used if none is provided in a call to its `execute <Mechanism_Base.execute>` or
        `run <Mechanism_Base.run>` methods; also serves as a template to specify the length of the `variable
        <DDM.variable>` for its `function <DDM.function>`, and the `primary OutputPort <OuputState_Primary>` of the
        DDM (see `Input` <DDM_Creation>` for how an input with a length of greater than 1 is handled).

    function : IntegratorFunction : default DriftDiffusionAnalytical
        specifies the function to use to `execute <DDM_Execution>` the decision process; determines the mode of
        execution (see `function <DDM.function>` and `DDM_Modes` for additional information).

    COMMENT:
    context=componentType+INITIALIZING):
        context : str : default ''None''
               string used for contextualization of instantiation, hierarchical calls, executions, etc.
    COMMENT

    Attributes
    ----------
    variable : value : default  FUNCTION_PARAMS[STARTING_VALUE]
        the input to Mechanism's execute method.  Serves as the "stimulus" component of the `function <DDM.function>`'s
        **drift_rate** parameter.

    function :  IntegratorFunction : default DriftDiffusionAnalytical
        the function used to `execute <DDM_Execution>` the decision process; determines the mode of execution.
        If it is `DriftDiffusionAnalytical <DriftDiffusionAnalytical>`, an `analytic solution
        <DDM_Analytic_Mode>` is calculated (note:  the latter requires that the MatLab engine is installed); if it is
        an `IntegratorFunction` Function with an `integration_type <IntegratorFunction.integration_type>` of
        *DIFFUSION*, then `numerical step-wise integration <DDM_Integration_Mode>` is carried out.  See `DDM_Modes` and
        `DDM_Execution` for additional information.
        COMMENT:
           IS THIS MORE CORRECT FOR ABOVE:
               if it is `DriftDiffusionIntegrator`, then `numerical step-wise integration <DDM_Integration_Mode>`
               is carried out.
        COMMENT

    value : 2d np.array[array(float64),array(float64),array(float64),array(float64)]
        result of executing DDM `function <DDM.function>`;  has six items, that are assigned based on the `function
        <DDM.function>` attribute.  The first two items are always assigned the values of `DECISION_VARIABLE
        <DDM_DECISION_VARIABLE>` and `RESPONSE_TIME <DDM_RESPONSE_TIME>` (though their interpretation depends on the
        `function <DDM.function>` and corresponding `mode of <DDM_Modes>` of operation).  See `DDM_Modes`,
        `DDM_Execution`, and `DDM_Output` for additional information about other values that can be reported and
        their interpretation.

    random_state : numpy.RandomState
        private pseudorandom number generator

    output_ports : ContentAddressableList[OutputPort]
        list of the DDM's `OutputPorts <OutputPort>`.  There are always two OutputPorts, `DECISION_VARIABLE
        <DDM_DECISION_VARIABLE>` and `RESPONSE_TIME <DDM_RESPONSE_TIME>`; additional ones may be included
        based on the `function <DDM.function>` and/or any specifications made in the **output_ports** argument of the
        DDM's constructor (see `DDM_Output` for additional details).

    output_values : List[array(float64)]
        each item is the `value <OutputPort.value>` of the corresponding `OutputPort` in `output_ports
        <DDM.output_ports>`.  The first two items are always the `value <OutputPort.value>`\\s of the
        `DECISION_VARIABLE <DDM_DECISION_VARIABLE>` and `RESPONSE_TIME <DDM_RESPONSE_TIME>` OutputPorts;  additional
        ones may be included, based on the `function <DDM.function>` and any specifications made in the
        **output_ports** argument of the DDM's constructor  (see `DDM_Output` for additional details).

    standard_output_ports : list[str]
        list of `Standard OutputPorts <OutputPort_Standard>` that includes the following addition to the
        `standard_output_ports <Mechanism_Base.standard_output_ports>` of a `Mechanism <Mechanism>`:

        .. _DDM_DECISION_VARIABLE:

        *DECISION_VARIABLE* : float
          • `analytic mode <DDM_Analytic_Mode>`: the value of the threshold crossed by the decision variable on the
            current TRIAL (which is either the value of the DDM `function <DDM.function>`'s threshold attribute or its
            negative); \n
          • `integration mode <DDM_Integration_Mode>`: the value of the decision variable at the current TIME_STEP of
            execution. \n
          Corresponds to the 1st item of the DDM's `value <DDM.value>`.

        .. _DDM_DECISION_VARIABLE_ARRAY:

        *DECISION_VARIABLE_ARRAY* : 1d nparray
          .. note::
             This is available only if **input_format** is specified as *ARRAY* in the DDM Mechanism's constructor
             (see `DDM_Input`).
          • `analytic mode <DDM_Analytic_Mode>`: two element array, with the decision variable (1st item of the DDM's
            `value <DDM.value>`) as the 1st element if the decision process crossed the upper threshold, and the 2nd element
            if it is closer to the lower threshold; the other element is set to 0. \n
          • `integration mode <DDM_Integration_Mode>`: the value of the decision variable at the current TIME_STEP of
            execution, assigned to the 1st element if the decision variable is closer to the upper threshold, and to the
            2nd element if it is closer to the lower threshold; the other element is set to 0. \n

        .. _DDM_SELECTED_INPUT_ARRAY:

        *SELECTED_INPUT_ARRAY* : 1d nparray
          .. note::
             This is available only if **input_format** is specified as *ARRAY* in the DDM Mechanism's constructor
             (see `DDM_Input`).
          • `analytic mode <DDM_Analytic_Mode>`: two element array, with one ("selected") element -- determined by the
            outcome of the decision process -- set to the value of the corresponding element in the stimulus array (i.e.,
            the DDM's input_port `variable <InputPort.variable>`).  The "selected" element is the 1st one if the decision
            process resulted in crossing the upper threshold, and the 2nd if it crossed the lower threshold; the other
            element is set to 0. \n
          • `integration mode <DDM_Integration_Mode>`: the value of the element in the stimulus array based on the
            decision variable (1st item of the DDM's `value <DDM.value>`) at the current TIME_STEP of execution:
            it is assigned to the 1st element if the decision variable is closer to the upper threshold, and to the
            2nd element if the decision variable is closer to the lower threshold; the other element is set to 0. \n

        .. _DDM_RESPONSE_TIME:

        *RESPONSE_TIME* : float
          • `analytic mode <DDM_Analytic_Mode>`: mean time (in seconds) for the decision variable to reach the positive
            or negative value of the DDM `function <DDM.function>`'s threshold attribute as estimated by the analytic
            solution calculated by the `function <DDM.function>`); \n
          • `integration mode <DDM_Integration_Mode>`: the number of `TIME_STEP`\\s that have occurred since the DDM
            began to execute in the current `TRIAL <TimeScale.TRIAL>` or, if it has reached the positive or negative
            value of the DDM `function <DDM.function>`'s threshold attribute, the `TIME_STEP` at which that occurred. \n
            Corresponds to the 2nd item of the DDM's `value <DDM.value>`.

        .. _DDM_DECISION_OUTCOME:

        *DECISION_OUTCOME* : float
          • `analytic mode <DDM_Analytic_Mode>`: 1.0 if the value of the threshold crossed by the decision variable on
            the current TRIAL (which is either the value of the DDM `function <DDM.function>`'s threshold attribute or
            its negative) is positive, 0 otherwise. \n
          • `integration mode <DDM_Integration_Mode>`: 1if the value of the decision variable at the current
            TIME_STEP of execution is positive, 0 otherwise. \n

        .. _DDM_PROBABILITY_UPPER_THRESHOLD:

        *PROBABILITY_UPPER_THRESHOLD* : float
          • `analytic mode <DDM_Analytic_Mode>`: the probability of the decision variable reaching the positive value of
            the DDM `function <DDM.function>`'s threshold attribute as estimated by the analytic solution calculated by the
            `function <DDM.function>`; often, by convention, the positive (upper) threshold is associated with the
            correct response, in which case *PROBABILITY_UPPER_THRESHOLD* corresponds to the accuracy of the decision
            process. \n
          • `integration mode <DDM_Integration_Mode>`: `None`.
          Corresponds to the 3rd item of the DDM's `value <DDM.value>`.

        COMMENT:
          [TBI:]
              `integration mode <DDM_Integration_Mode>`, if execution has completed, this is a binary value
              indicating whether the decision process reached the upper (positive) threshold. If execution was
              interrupted (using :py:meth:`terminate_function  <DDM.terminate_function>`, sometimes referred to as the
              :ref:`interrogation protocol <LINK>`, then the value corresponds to the current likelihood that the upper
              threshold would have been reached.
        COMMENT

        .. _DDM_PROBABILITY_LOWER_THRESHOLD:

        *PROBABILITY_LOWER_THRESHOLD* : float
          • `analytic mode <DDM_Analytic_Mode>`: the probability of the decision variable reaching the negative value of
            the DDM `function <DDM.function>`'s threshold attribute as estimated by the analytic solution calculate by the
            `function <DDM.function>`); often, by convention, the negative (lower) threshold is associated with an error
            response, in which case *PROBABILITY_LOWER_THRESHOLD* corresponds to the error rate of the decision process; \n
          • `integration mode <DDM_Integration_Mode>`: `None`.
          Corresponds to the 4th item of the DDM's `value <DDM.value>`.

            COMMENT:
              [TBI:]
                  `integration mode <DDM_Integration_Mode>`, if execution has completed, this is a binary value
                  indicating whether the decision process reached the lower (negative) threshold. If execution was
                  interrupted (using :py:method:`terminate_method <DDM.terminate_function>`, sometimes referred to as the
                  :ref:`interrogation protocol <LINK>`), then the value corresponds to the current likelihood that the lower
                  threshold would have been reached.
            COMMENT

        .. _DDM_RT_CORRECT_MEAN:

        *RT_CORRECT_MEAN* : floa
          (only applicable if `function <DDM.function>` is `DriftDiffusionAnalytical`) \n
          • `analytic mode <DDM_Analytic_Mode>`:  the mean reaction time (in seconds) for responses in which the decision
            variable reached the positive value of the DDM `function <DDM.function>`'s threshold attribute as estimated by
            closed form analytic solutions from Srivastava et al. (https://arxiv.org/abs/1601.06420) \n
          • `integration mode <DDM_Integration_Mode>`: `None`.
          Corresponds to the 5th item of the DDM's `value <DDM.value>`.

        .. _DDM_RT_CORRECT_VARIANCE:

        *RT_CORRECT_VARIANCE* : float
          (only applicable if `function <DDM.function>` is `DriftDiffusionAnalytical`) \n
          • `analytic mode <DDM_Analytic_Mode>`:  the variance of reaction time (in seconds) for responses in which the
            decision variable reached the positive value of the DDM `function <DDM.function>`'s threshold attribute as
            estimated by closed form analytic solutions from Srivastava et al. (https://arxiv.org/abs/1601.06420) \n
          • `integration mode <DDM_Integration_Mode>`: `None`.
          Corresponds to the 6th item of the DDM's `value <DDM.value>`.

        .. _DDM_RT_CORRECT_SKEW:

        *RT_CORRECT_SKEW* : float
          (only applicable if `function <DDM.function>` is `DriftDiffusionAnalytical`) \n
          • `analytic mode <DDM_Analytic_Mode>`:  the skew of decision time (in seconds) for responses in which the decision
            variable reached the positive value of the DDM `function <DDM.function>`'s threshold attribute as estimated by
            closed form analytic solutions from Srivastava et al. (https://arxiv.org/abs/1601.06420) \n
          • `integration mode <DDM_Integration_Mode>`: `None`.
          Corresponds to the 7th item of the DDM's `value <DDM.value>`.

        .. _DDM_RT_INCORRECT_MEAN:

        *RT_INCORRECT_MEAN* : float
          (only applicable if `function <DDM.function>` is `DriftDiffusionAnalytical`) \n
          • `analytic mode <DDM_Analytic_Mode>`:  the mean reaction time (in seconds) for responses in which the decision
            variable reached the negative value of the DDM `function <DDM.function>`'s threshold attribute as estimated by
            closed form analytic solutions from Srivastava et al. (https://arxiv.org/abs/1601.06420) \n
          • `integration mode <DDM_Integration_Mode>`: `None`.
          Corresponds to the 5th item of the DDM's `value <DDM.value>`.

        .. _DDM_RT_INCORRECT_VARIANCE:

        *RT_INCORRECT_VARIANCE* : float
          (only applicable if `function <DDM.function>` is `DriftDiffusionAnalytical`) \n
          • `analytic mode <DDM_Analytic_Mode>`:  the variance of reaction time (in seconds) for responses in which the
            decision variable reached the negative value of the DDM `function <DDM.function>`'s threshold attribute as
            estimated by closed form analytic solutions from Srivastava et al. (https://arxiv.org/abs/1601.06420) \n
          • `integration mode <DDM_Integration_Mode>`: `None`.
          Corresponds to the 6th item of the DDM's `value <DDM.value>`.

        .. _DDM_RT_INCORRECT_SKEW:

        *RT_INCORRECT_SKEW* : float
          (only applicable if `function <DDM.function>` is `DriftDiffusionAnalytical`) \n
          • `analytic mode <DDM_Analytic_Mode>`:  the skew of decision time (in seconds) for responses in which the decision
            variable reached the negative value of the DDM `function <DDM.function>`'s threshold attribute as estimated by
            closed form analytic solutions from Srivastava et al. (https://arxiv.org/abs/1601.06420) \n
          • `integration mode <DDM_Integration_Mode>`: `None`.
          Corresponds to the 7th item of the DDM's `value <DDM.value>`.

    """

    componentType = "DDM"

    classPreferenceLevel = PreferenceLevel.SUBTYPE
    # These will override those specified in SUBTYPE_DEFAULT_PREFERENCES
    classPreferences = {
        PREFERENCE_SET_NAME: 'DDMCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    class Parameters(ProcessingMechanism.Parameters):
        """
            Attributes
            ----------

                function
                    see `function <DDM.function>`

                    :default value: `DriftDiffusionAnalytical`
                    :type: `Function`

                initializer
                    see `initializer <DDM.initializer>`

                    :default value: numpy.array([[0]])
                    :type: ``numpy.ndarray``

                input_format
                    see `input_format <DDM.input_format>`

                    :default value: `SCALAR`
                    :type: ``str``

                random_state
                    see `random_state <DDM.random_state>`

                    :default value: None
                    :type: ``numpy.random.RandomState``
        """
        function = Parameter(
            DriftDiffusionAnalytical(
                drift_rate=1.0,
                starting_value=0.0,
                threshold=1.0,
                noise=0.5,
                non_decision_time=.200,
            ),
            stateful=False,
            loggable=False
        )
        input_format = Parameter(SCALAR, stateful=False, loggable=False)
        initializer = np.array([[0]])
        random_state = Parameter(None, loggable=False, getter=_random_state_getter, dependencies='seed')
        seed = Parameter(DEFAULT_SEED(), modulable=True, fallback_default=True, setter=_seed_setter)

        output_ports = Parameter(
            [DECISION_VARIABLE, RESPONSE_TIME],
            stateful=False,
            loggable=False,
            read_only=True,
            structural=True,
        )

    standard_output_ports =[{NAME: DECISION_VARIABLE,},           # Upper or lower threshold for Analtyic function
                            {NAME: RESPONSE_TIME},                # TIME_STEP within TRIAL for Integrator function
                            {NAME: PROBABILITY_UPPER_THRESHOLD},  # Accuracy (TRIAL mode only)
                            {NAME: PROBABILITY_LOWER_THRESHOLD},  # Error rate (TRIAL mode only)
                            {NAME: RT_CORRECT_MEAN},              # (DriftDiffusionAnalytical only)
                            {NAME: RT_CORRECT_VARIANCE},          # (DriftDiffusionAnalytical only)
                            {NAME: RT_CORRECT_SKEW},              # (DriftDiffusionAnalytical only)
                            {NAME: RT_INCORRECT_MEAN},            # (DriftDiffusionAnalytical only)
                            {NAME: RT_INCORRECT_VARIANCE},        # (DriftDiffusionAnalytical only)
                            {NAME: RT_INCORRECT_SKEW},            # (DriftDiffusionAnalytical only)
                            ]
    standard_output_port_names = [i['name'] for i in standard_output_ports]

    @check_user_specified
    @beartype
    def __init__(self,
                 default_variable=None,
                 input_shapes=None,
                 input_format: Optional[Literal['SCALAR', 'ARRAY', 'VECTOR']] = None,
                 function=None,
                 input_ports=None,
                 output_ports: Optional[Union[str, Iterable]] = None,
                 seed=None,
                 params=None,
                 name=None,
                 prefs:  Optional[ValidPrefSet] = None,
                 **kwargs):

        # Override instantiation of StandardOutputPorts usually done in _instantiate_output_ports
        #    in order to use SEQUENTIAL indices
        self.standard_output_ports = StandardOutputPorts(self, self.standard_output_ports, indices=SEQUENTIAL)

        if input_format is not None and input_ports is not None:
            raise DDMError(
                'Only one of input_format and input_ports should be specified.'
            )
        elif input_format is None:
            input_format = SCALAR

        # If input_format is specified to be ARRAY or VECTOR, instantiate:
        #    InputPort with:
        #        2-item array as its variable
        #        Reduce as its function, which will generate an array of len 1
        #        and therefore specify size of Mechanism's variable as 1
        #    OutputPorts that report the decision variable and selected input in array format
        #        IMPLEMENTATION NOTE:
        #            These are created here rather than as StandardOutputPorts
        #            since they require input_format==ARRAY to be meaningful
        if input_format in {ARRAY, VECTOR}:
            input_shapes=1 # size of variable for DDM Mechanism
            input_ports = [
                {NAME:'ARRAY',
                 VARIABLE: np.array([[0.0, 0.0]]),
                 FUNCTION: Reduce(weights=[1,-1])}
            ]
            self.standard_output_ports.add_port_dicts([
                # Provides a 1d 2-item array with:
                #    decision variable in position corresponding to threshold crossed, and 0 in the other position
                {NAME: DECISION_VARIABLE_ARRAY, # 1d len 2, DECISION_VARIABLE as element 0 or 1
                 VARIABLE:[(OWNER_VALUE, self.DECISION_VARIABLE_INDEX), THRESHOLD],
                           # per VARIABLE assignment above, items of v of lambda function below are:
                           #    v[0]=self.value[self.DECISION_VARIABLE_INDEX]
                           #    v[1]=self.parameter_ports[THRESHOLD]
                 FUNCTION: lambda v: [float(v[0]), 0] if (v[1] - v[0]) < (v[1] + v[0]) else [0, float(v[0])]},
                # Provides a 1d 2-item array with:
                #    input value in position corresponding to threshold crossed by decision variable, and 0 in the other
                {NAME: SELECTED_INPUT_ARRAY, # 1d len 2, DECISION_VARIABLE as element 0 or 1
                 VARIABLE:[(OWNER_VALUE, self.DECISION_VARIABLE_INDEX), THRESHOLD, (INPUT_PORT_VARIABLES, 0)],
                 # per VARIABLE assignment above, items of v of lambda function below are:
                 #    v[0]=self.value[self.DECISION_VARIABLE_INDEX]
                 #    v[1]=self.parameter_ports[THRESHOLD]
                 #    v[2]=self.input_ports[0].variable
                 FUNCTION: lambda v: [float(v[2][0][0]), 0]
                                      if (v[1] - v[0]) < (v[1] + v[0])
                                      else [0, float(v[2][0][1])]
                 }
            ])

        self.standard_output_ports.add_port_dicts([
            {
                NAME: DECISION_OUTCOME,
                VARIABLE: (OWNER_VALUE, self.DECISION_VARIABLE_INDEX),
                FUNCTION: UserDefinedFunction(custom_function=lambda x: (x > 0.0).astype(float))
            }
        ])

        # Add StandardOutputPorts for Mechanism (after ones for DDM, so that their indices are not messed up)
        # FIX 11/9/19:  ADD BACK ONCE Mechanism_Base.standard_output_ports ONLY HAS RESULTS IN ITS
        # self.standard_output_ports.add_port_dicts(Mechanism_Base.standard_output_ports)

        # Default output_ports is specified in constructor as a tuple rather than a list
        # to avoid "gotcha" associated with mutable default arguments
        # (see: bit.ly/2uID3s3 and http://docs.python-guide.org/en/latest/writing/gotchas/)
        if isinstance(output_ports, (str, tuple)):
            output_ports = list(output_ports)

        # IMPLEMENTATION NOTE: this manner of setting default_variable works but is idiosyncratic
        # compared to other mechanisms: see TransferMechanism.py __init__ function for a more normal example.
        if default_variable is None and input_shapes is None:
            try:
                default_variable = params[FUNCTION_PARAMS][STARTING_VALUE]
                if not is_numeric(default_variable):
                    # set normally by default
                    default_variable = None
            except (KeyError, TypeError):
                # set normally by default
                pass

        # # Conflict with above
        # self.input_shapes = input_shapes

        # New (1/19/2021) default behaviour of DDM mechanism is to execute until finished. That
        # is, it should execute until it reaches its threshold.
        self.parameters.execute_until_finished.default_value = True

        # FIX: MAKE THIS A PARAMETER WITH ARGUMENT IN CONSTRUCTOR (ALIGN WITH reset Parameter OF IntegratorMechanism)
        # New (1/19/2021) default behavior of DDM mechanism is to reset stateful functions
        # on each new trial.
        if 'reset_stateful_function_when' not in kwargs:
            kwargs['reset_stateful_function_when'] = AtTrialStart()

        # Set maximum executions absurdly large to avoid early termination
        self.max_executions_before_finished = sys.maxsize

        super(DDM, self).__init__(default_variable=default_variable,
                                  seed=seed,
                                  input_ports=input_ports,
                                  output_ports=output_ports,
                                  function=function,
                                  params=params,
                                  name=name,
                                  prefs=prefs,
                                  input_shapes=input_shapes,
                                  **kwargs),

        self._instantiate_plotting_functions()

    def plot(self, stimulus=1.0, threshold=10.0):
        """
        Generate a dynamic plot of the DDM integrating over time towards a threshold.

        .. note::
            The plot method is only available when the DriftDiffusionIntegrator function is in use. The plot method does
            not represent the results of this DDM mechanism in particular, and does not affect the current state of this
            mechanism's DriftDiffusionIntegrator. The plot method is only meant to visualize a possible path of a DDM
            mechanism with these function parameters.

        Arguments
        ---------
        stimulus: float: default 1.0
            specify a stimulus value for the AdaptiveIntegrator function

        threshold: float: default 10.0
            specify the threshold at which the DDM will stop integrating

        Returns
        -------
        Mechanism's function plot : Matplotlib window
            Matplotlib window of the Mechanism's function plotting dynamically over time with specified parameters
            towards a specified threshold


        """
        import matplotlib.pyplot as plt
        import time
        plt.ion()

        # set initial values and threshold
        time_step = [0]
        position = [float(self.defaults.variable)]
        variable = stimulus

        # execute the mechanism once to begin the loop
        result_check = self.plot_function(variable, context="plot")[0][0]

        # continue executing the ddm until its value exceeds the threshold
        while abs(result_check) < threshold:
            time_step.append(time_step[-1] + 1)
            position.append(result_check)
            result_check = self.plot_function(variable, context="plot")[0][0]

        # add the ddm's final position to the list of positions
        time_step.append(time_step[-1] + 1)
        position.append(result_check)

        figure, ax = plt.subplots(1, 1)
        lines, = ax.plot([], [], 'o')
        ax.set_xlim(0, time_step[-1])
        ax.set_ylim(-threshold, threshold)
        ax.grid()
        xdata = []
        ydata = []

        # add each of the position values to the plot one at a time
        for t in range(time_step[-1]):
            xdata.append(t)
            ydata.append(position[t])
            lines.set_xdata(xdata)
            lines.set_ydata(ydata)
            figure.canvas.draw()
            # number of seconds to wait before next point is plotted
            time.sleep(.1)

    def _validate_variable(self, variable, context=None):
        """Ensures that input to DDM is a single value.
        Remove when MULTIPROCESS DDM is implemented.
        """

        # this test may become obsolete when size is moved to Component.py
        # if len(variable) > 1 and not self.input_format in {ARRAY, VECTOR}:
        if not object_has_single_value(variable) and not object_has_single_value(np.array(variable)):
            raise DDMError("Length of input to DDM ({}) is greater than 1, implying there are multiple "
                           "input ports, which is currently not supported in DDM, but may be supported"
                           " in the future under a multi-process DDM. Please use a single numeric "
                           "item as the default_variable, or use input_shapes = 1.".format(variable))
        # # MODIFIED 6/28/17 (CW): changed len(variable) > 1 to len(variable[0]) > 1
        # # if not isinstance(variable, numbers.Number) and len(variable[0]) > 1:
        # if not is_numeric(variable) and len(variable[0]) > 1:
        #     raise DDMError("Input to DDM ({}) must have only a single numeric item".format(variable))
        return super()._validate_variable(variable=variable, context=context)

    def _validate_params(self, request_set, target_set=None, context=None):

        super()._validate_params(request_set=request_set, target_set=target_set, context=context)
        functions = {DriftDiffusionAnalytical,
                     DriftDiffusionIntegrator}

        if FUNCTION in target_set and target_set[FUNCTION] is not None:
            # If target_set[FUNCTION] is a method of a Function (e.g., being assigned in _instantiate_function),
            #   get the Function to which it belongs
            fun = target_set[FUNCTION]
            if isinstance(fun, types.MethodType):
                fun = fun.__self__.__class__

            for function_type in functions:
                if is_same_function_spec(fun, function_type):
                    break
            else:
                function_names = [fun.componentName for fun in functions]
                raise DDMError("{} param of {} must be one of the following functions: {}".
                               format(FUNCTION, self.name, function_names))

        try:
            threshold = target_set[FUNCTION_PARAMS][THRESHOLD]
        except KeyError:
            pass
        else:
            if isinstance(threshold, tuple):
                threshold = threshold[0]
            if is_numeric(threshold):
                if not threshold >= 0:
                    raise DDMError("{} param of {} ({}) must be >= zero".
                                   format(THRESHOLD, self.name, threshold))
            elif isinstance(threshold, ControlSignal):
                threshold = threshold.allocation_samples
                if not np.amin(threshold) >= 0:
                    raise DDMError("The lowest value of {} for the {} "
                                   "assigned to the {} param of {} must be >= zero".
                                   format(ALLOCATION_SAMPLES, ControlSignal.__name__, THRESHOLD, self.name, threshold))
            elif _is_control_spec(threshold):
                pass
            else:
                raise DDMError("PROGRAM ERROR: unrecognized specification for {} of {} ({})".
                               format(THRESHOLD, self.name, threshold))

    def _instantiate_plotting_functions(self, context=None):
        if "DriftDiffusionIntegrator" in str(self.function):
            self.get_axes_function = DriftDiffusionIntegrator(
                rate=self.function.defaults.rate,
                noise=self.function.defaults.noise
            ).function
            self.plot_function = DriftDiffusionIntegrator(
                rate=self.function.defaults.rate,
                noise=self.function.defaults.noise
            ).function

    def _execute(
        self,
        variable=None,
        context=None,
        runtime_params=None,

    ):
        """Execute DDM function (currently only trial-level, analytic solution)
        Execute DDM and estimate outcome or calculate trajectory of decision variable
        Currently implements only trial-level DDM (analytic solution) and returns:
            - stochastically estimated decion outcome (convert mean ER into value between 1 and -1)
            - mean ER
            - mean RT
            - mean, variance, and skew of RT for correct (posititive threshold) responses
            - mean, variance, and skew of RT for incorrect (negative threshold) responses
        Return current decision variable (self.outputPort.value) and other output values (self.output_ports[].value
        Arguments:
        # CONFIRM:
        variable (float): set to self.value (= self.input_value)
        - params (dict):  runtime_params passed from Mechanism, used as one-time value for current execution:
            + DRIFT_RATE (float)
            + THRESHOLD (float)
            + kwDDM_Bias (float)
            + NON_DECISION_TIME (float)
            + NOISE (float)
        - context (str)
        Returns the following values in self.value (2D np.array) and in
            the value of the corresponding outputPort in the self.output_ports dict:
            - decision variable (float)
            - mean error rate (float)
            - mean RT (float)
            - mean, variance, and skew of RT for correct (posititive threshold) responses
            - mean, variance, and skew of RT for incorrect (negative threshold) responses

        :param self:
        :param variable (float)
        :param params: (dict)
        :param context: (str)
        :rtype self.outputPort.value: (number)
        """

        if variable is None or any(np.isnan(i) for i in variable):
            # IMPLEMENT: MULTIPROCESS DDM:  ??NEED TO DEAL WITH PARTIAL NANS
            variable = self.defaults.variable

        variable = self._validate_variable(variable)

        # EXECUTE INTEGRATOR SOLUTION (TIME_STEP TIME SCALE) -----------------------------------------------------
        if isinstance(self.function, IntegratorFunction):

            result = super()._execute(variable, context=context)

            if self.initialization_status != ContextFlags.INITIALIZING:
                logger.info('{0} {1} is at {2}'.format(type(self).__name__, self.name, result))

            return convert_all_elements_to_np_array([result[0], result[1]])

        # EXECUTE ANALYTIC SOLUTION (TRIAL TIME SCALE) -----------------------------------------------------------
        else:

            result = super()._execute(
                variable=variable,
                context=context,
                runtime_params=runtime_params,
            )

            if isinstance(self.function, DriftDiffusionAnalytical):
                return_value = np.zeros(shape=(10,1))
                return_value[self.RESPONSE_TIME_INDEX] = result[0]
                return_value[self.PROBABILITY_LOWER_THRESHOLD_INDEX] = result[1]
                return_value[self.PROBABILITY_UPPER_THRESHOLD_INDEX] = \
                                                               1 - return_value[self.PROBABILITY_LOWER_THRESHOLD_INDEX]
                return_value[self.RT_CORRECT_MEAN_INDEX] = result[2]
                return_value[self.RT_CORRECT_VARIANCE_INDEX] = result[3]
                return_value[self.RT_CORRECT_SKEW_INDEX] = result[4]
                return_value[self.RT_INCORRECT_MEAN_INDEX] = result[5]
                return_value[self.RT_INCORRECT_VARIANCE_INDEX] = result[6]
                return_value[self.RT_INCORRECT_SKEW_INDEX] = result[7]

            else:
                raise DDMError("The function specified ({}) for {} is not a valid function selection for the DDM".
                               format(self.function.name, self.name))

            # Convert ER to decision variable:
            threshold = float(self.function._get_current_parameter_value(THRESHOLD, context).item())
            random_state = self._get_current_parameter_value(self.parameters.random_state, context)
            if random_state.uniform() < return_value[self.PROBABILITY_LOWER_THRESHOLD_INDEX]:
                return_value[self.DECISION_VARIABLE_INDEX] = np.atleast_1d(-1 * threshold)
            else:
                return_value[self.DECISION_VARIABLE_INDEX] = threshold
            return return_value

    def _gen_llvm_invoke_function(self, ctx, builder, function, params, state,
                                  variable, m_val, *, tags:frozenset):

        if isinstance(self.function, IntegratorFunction):
            # Integrator based DDM works like other mechanisms
            return super()._gen_llvm_invoke_function(ctx, builder, function,
                                                     params, state, variable,
                                                     m_val, tags=tags)

        elif isinstance(self.function, DriftDiffusionAnalytical):
            mf_out, builder = super()._gen_llvm_invoke_function(ctx, builder, function,
                                                                params, state, variable,
                                                                None, tags=tags)
            # The order and number of returned values is different for DDA
            for res_idx, idx in enumerate((self.RESPONSE_TIME_INDEX,
                                           self.PROBABILITY_LOWER_THRESHOLD_INDEX,
                                           self.RT_CORRECT_MEAN_INDEX,
                                           self.RT_CORRECT_VARIANCE_INDEX,
                                           self.RT_CORRECT_SKEW_INDEX,
                                           self.RT_INCORRECT_MEAN_INDEX,
                                           self.RT_INCORRECT_VARIANCE_INDEX,
                                           self.RT_INCORRECT_SKEW_INDEX)):
                src = builder.gep(mf_out, [ctx.int32_ty(0), ctx.int32_ty(res_idx)])
                dst = builder.gep(m_val, [ctx.int32_ty(0), ctx.int32_ty(idx)])
                builder.store(builder.load(src), dst)

            # Handle upper threshold probability (1 - Lower Threshold)
            src = builder.gep(m_val, [ctx.int32_ty(0),
                                      ctx.int32_ty(self.PROBABILITY_LOWER_THRESHOLD_INDEX),
                                      ctx.int32_ty(0)])
            dst = builder.gep(m_val, [ctx.int32_ty(0),
                                      ctx.int32_ty(self.PROBABILITY_UPPER_THRESHOLD_INDEX),
                                      ctx.int32_ty(0)])
            prob_lower_thr = builder.load(src)
            prob_upper_thr = builder.fsub(prob_lower_thr.type(1), prob_lower_thr)
            builder.store(prob_upper_thr, dst)

            # Store threshold as decision variable output
            # this will be used by the mechanism to return the right decision
            threshold_ptr = ctx.get_param_or_state_ptr(builder,
                                                       self.function,
                                                       THRESHOLD,
                                                       param_struct_ptr=params)
            threshold = pnlvm.helpers.load_extract_scalar_array_one(builder, threshold_ptr)
            decision_ptr = builder.gep(m_val, [ctx.int32_ty(0),
                                               ctx.int32_ty(self.DECISION_VARIABLE_INDEX),
                                               ctx.int32_ty(0)])
            builder.store(threshold, decision_ptr)
        else:
            assert False, "Unknown mode in compiled DDM!"

        return m_val, builder

    def _gen_llvm_mechanism_functions(self, ctx, builder, m_base_params, m_params, m_state, m_in,
                                      m_val, ip_output, *, tags:frozenset):

        mf_out, builder = super()._gen_llvm_mechanism_functions(ctx, builder, m_base_params,
                                                                m_params, m_state, m_in, m_val,
                                                                ip_output, tags=tags)
        assert mf_out is m_val

        if isinstance(self.function, DriftDiffusionAnalytical):
            random_state = ctx.get_random_state_ptr(builder, self, m_state, m_params)
            random_f = ctx.get_uniform_dist_function_by_state(random_state)
            random_val_ptr = builder.alloca(random_f.args[1].type.pointee, name="random_out")
            builder.call(random_f, [random_state, random_val_ptr])
            random_val = builder.load(random_val_ptr)

            # Convert ER to decision variable:
            prob_lthr_ptr = builder.gep(m_val, [ctx.int32_ty(0),
                                                ctx.int32_ty(self.PROBABILITY_LOWER_THRESHOLD_INDEX),
                                                ctx.int32_ty(0)])
            prob_lower_thr = builder.load(prob_lthr_ptr)
            thr_cmp = builder.fcmp_ordered("<", random_val, prob_lower_thr)

            # The correct (modulated) threshold value is passed as
            # decision variable output
            decision_ptr = builder.gep(m_val, [ctx.int32_ty(0),
                                               ctx.int32_ty(self.DECISION_VARIABLE_INDEX),
                                               ctx.int32_ty(0)])
            threshold = builder.load(decision_ptr)
            neg_threshold = pnlvm.helpers.fneg(builder, threshold)
            res = builder.select(thr_cmp, neg_threshold, threshold)

            builder.store(res, decision_ptr)

        return m_val, builder

    @handle_external_context(fallback_most_recent=True)
    def reset(self, *args, force=False, context=None, **kwargs):
        from psyneulink.core.components.functions.stateful.integratorfunctions import IntegratorFunction

        # (1) reset function, (2) update mechanism value, (3) update output ports
        if isinstance(self.function, IntegratorFunction):
            new_values = self.function.reset(*args, **kwargs, context=context)
            self.parameters.value._set(convert_all_elements_to_np_array(new_values), context)
            self._update_output_ports(context=context)

    @handle_external_context()
    def is_finished(self, context=None):
        # find the single numeric entry in previous_value
        try:
            single_value = self.function.parameters.previous_value._get(context)
        except AttributeError:
            # Analytical function so it is always finished after it is called
            return True

        # indexing into a matrix doesn't reduce dimensionality
        if not isinstance(single_value, (np.matrix, str)):
            while True:
                try:
                    single_value = single_value[0]
                except (IndexError, TypeError):
                    break

        if (
            abs(single_value) >= self.function._get_current_parameter_value(THRESHOLD, context)
            and isinstance(self.function, IntegratorFunction)
        ):
            logger.info(
                '{0} {1} has reached threshold {2}'.format(
                    type(self).__name__,
                    self.name,
                    self.function._get_current_parameter_value(THRESHOLD, context)
                )
            )
            return True
        return False

    def _gen_llvm_is_finished_cond(self, ctx, builder, m_base_params, m_state, m_in):
        # Setup pointers to internal function
        f_base_params, f_state = ctx.get_param_or_state_ptr(builder,
                                                            self,
                                                            "function",
                                                            param_struct_ptr=m_base_params,
                                                            state_struct_ptr=m_state)

        # Find the single numeric entry in previous_value.
        # This exists only if the 'function' is 'integrator'
        prev_val_ptr = ctx.get_param_or_state_ptr(builder, self.function, "previous_value", state_struct_ptr=f_state)
        if prev_val_ptr is None:
            return ctx.bool_ty(1)

        # Extract scalar value from ptr
        prev_val_ptr = builder.gep(prev_val_ptr, [ctx.int32_ty(0), ctx.int32_ty(0)])
        prev_val = builder.load(prev_val_ptr)

        # Take abs of previous val
        llvm_fabs = ctx.get_builtin("fabs", [ctx.float_ty])
        prev_val = builder.call(llvm_fabs, [prev_val])

        # Get functions params and apply modulation
        f_params, builder = self._gen_llvm_param_ports_for_obj(
                self.function, f_base_params, ctx, builder, m_base_params, m_state, m_in)

        # Get threshold value
        threshold_ptr = ctx.get_param_or_state_ptr(builder,
                                                   self.function,
                                                   THRESHOLD,
                                                   param_struct_ptr=f_params)

        threshold_ptr = builder.gep(threshold_ptr, [ctx.int32_ty(0), ctx.int32_ty(0)])
        threshold = builder.load(threshold_ptr)
        is_prev_greater_or_equal = builder.fcmp_ordered('>=', prev_val, threshold)

        return is_prev_greater_or_equal



import psyneulink as pnl
import pytest


def pytest_assertrepr_compare(op, left, right):
    if isinstance(left, list) and isinstance(right, list) and op == '==':
        return [
            'Time Step output matching:',
            'Actual output:', str(left),
            'Expected output:', str(right)
        ]


@pytest.helpers.register
def setify_expected_output(expected_output):
    type_set = type(set())
    for i in range(len(expected_output)):
        if type(expected_output[i]) is not type_set:
            try:
                iter(expected_output[i])
                expected_output[i] = set(expected_output[i])
            except TypeError:
                expected_output[i] = set([expected_output[i]])
    return expected_output


@pytest.fixture
def three_node_linear_composition():
    A = pnl.TransferMechanism(name='A')
    B = pnl.TransferMechanism(name='B')
    C = pnl.TransferMechanism(name='C')

    comp = pnl.Composition()
    comp.add_linear_processing_pathway([A, B, C])

    return comp.nodes, comp


@pytest.helpers.register
def composition_from_string_pathways(pathways):
    mechanisms = {}
    pathways_as_mechs = []

    for p in pathways:
        p_as_mechs = []
        assert not isinstance(p, str), 'pathways must be a list of lists'
        for m in p:
            try:
                mech = mechanisms[m]
            except KeyError:
                mech = pnl.ProcessingMechanism(name=m)
                mechanisms[m] = mech
            p_as_mechs.append(mech)
        pathways_as_mechs.append(p_as_mechs)

    comp = pnl.Composition(pathways=pathways_as_mechs)

    return comp, mechanisms, mechanisms.values()
